# -*- coding: utf-8 -*-
"""
Encoding of the RDFa prototype vocabulary behavior. This means processing the graph by adding and removing triples
based on triples using the rdfa:Prototype and rdfa:ref class and property, respectively. For details, see the HTML5+RDFa document.


@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
@contact: Ivan Herman, ivan@w3.org
@version: $Id: prototype.py,v 1.1 2013-01-18 09:41:49 ivan Exp $
$Date: 2013-01-18 09:41:49 $
"""

from rdflib import RDF as ns_rdf

from .. import ns_rdfa

Prototype = ns_rdfa["Pattern"]
pref = ns_rdfa["copy"]

def handle_prototypes(graph):
    to_remove = set()
    for x, ref, PR in graph.triples((None, pref, None)):
        if (PR, ns_rdf["type"], Prototype) in graph:
            to_remove.add((PR, ns_rdf["type"], Prototype))
            to_remove.add((x, ref, PR))
            # there is a reference to a prototype here
            for (PR, p, y) in graph.triples((PR, None, None)):
                if not (p == ns_rdf["type"] and y == Prototype):
                    graph.add((x, p, y))
                    to_remove.add((PR, p, y))
    for t in to_remove: graph.remove(t)




# -*- coding: utf-8 -*-
"""
Simple transfomer for HTML5: add a @src for any @data, add a @content for the @value attribute of the <data> element, and interpret the <time> element.

@summary: Add a top "about" to <head> and <body>
@requires: U{RDFLib package<http://rdflib.net>}
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
@contact: Ivan Herman, ivan@w3.org
"""

"""
$Id: html5.py,v 1.15 2014-11-04 13:18:48 ivan Exp $
$Date: 2014-11-04 13:18:48 $
"""

from functools import reduce

# The handling of datatime is a little bit more complex... better put this in a separate function for a better management
from datetime import datetime
import re

datetime_type =   "http://www.w3.org/2001/XMLSchema#dateTime"
time_type =       "http://www.w3.org/2001/XMLSchema#time"
date_type =       "http://www.w3.org/2001/XMLSchema#date"
date_gYear =      "http://www.w3.org/2001/XMLSchema#gYear"
date_gYearMonth = "http://www.w3.org/2001/XMLSchema#gYearMonth"
date_gMonthDay =  "http://www.w3.org/2001/XMLSchema#gMonthDay"
duration_type =   "http://www.w3.org/2001/XMLSchema#duration"
plain =           "plain"

handled_time_types = [datetime_type, time_type, date_type, date_gYear, date_gYearMonth, date_gMonthDay, duration_type]

_formats = {
    date_gMonthDay:  ["%m-%d"],
    date_gYearMonth: ["%Y-%m"],
    date_gYear:      ["%Y"],
    date_type:       ["%Y-%m-%d", "%Y-%m-%dZ"],
    time_type:       ["%H:%M",
                      "%H:%M:%S",
                      "%H:%M:%SZ",
                      "%H:%M:%S.%f"],
    datetime_type:   ["%Y-%m-%dT%H:%M",
                      "%Y-%m-%dT%H:%M:%S",
                      "%Y-%m-%dT%H:%M:%S.%f",
                      "%Y-%m-%dT%H:%MZ",
                      "%Y-%m-%dT%H:%M:%SZ",
                      "%Y-%m-%dT%H:%M:%S.%fZ"],
    duration_type:   ["P%dD",
                      "P%YY%mM%dD",
                      "P%YY%mM",
                      "P%YY%dD",
                      "P%YY",
                      "P%mM",
                      "P%mM%dD"]
}

_dur_times = ["%HH%MM%SS", "%HH", "%MM", "%SS", "%HH%MM", "%HH%SS", "%MM%SS"]

def _format_test(string):
    """
    Tests the string format to see whether it fits one of the time datatypes
    @param string: attribute value to test
    @return: a URI for the xsd datatype or the string 'plain'
    """
    # Try to get the easy cases:
    for key in _formats:
        for f in _formats[key]:
            try :
                # try to check if the syntax is fine
                _d = datetime.strptime(string, f)
                # bingo!
                return key
            except ValueError:
                pass
            
    # Now come the special cases:-(
    # Check first for the duration stuff, that is the nastiest.
    if len(string) > 2 and (string[0] == 'P' or (string [0] == '-' and string[1] == 'P')):
        # this is meant to be a duration type
        # first of all, get rid of the leading '-' and check again
        if string[0] == '-':
            for f in _formats[duration_type]:
                try :
                    # try to check if the syntax is fine
                    _d = datetime.strptime(string, f)
                    # bingo!
                    return duration_type
                except ValueError :
                    pass
        # Let us see if the value contains a separate time portion, and cut that one
        durs = string.split('T')
        if len(durs) == 2 :
            # yep, so we should check again
            dur = durs[0]
            tm  = durs[1]
            # Check the duration part
            td = False
            for f in _formats[duration_type] :
                try :
                    # try to check if the syntax is fine
                    _d = datetime.strptime(dur, f)
                    # bingo!
                    td = True
                    break
                except ValueError :
                    pass
            if td == True :
                # Getting there...
                for f in _dur_times :
                    try :
                        # try to check if the syntax is fine
                        _d = datetime.strptime(tm, f)
                        # bingo!
                        return duration_type
                    except ValueError :
                        pass
            # something went wrong
            return plain
        else:
            # Well, no more tricks, this is a plain type
            return plain

    # If we got here, we should check the time zone
    # there is a discrepancy betwen the python and the HTML5/XSD lexical string,
    # which means that this has to handled separately for the date and the timezone portion
    try:
        # The time-zone-less portion of the string
        s = string[0:-6]
        # The time-zone portion
        tz = string[-5:]
        try :
            _t = datetime.strptime(tz,"%H:%M")
        except ValueError :
            # Bummer, this is not a correct time
            return plain
        # The time-zone is fine, the datetime portion has to be checked        
        for f in _formats[datetime_type] :
            try :
                # try to check if it is fine
                _d = datetime.strptime(s, f)
                # Bingo!
                return datetime_type
            except ValueError :
                pass
    except :
        pass
    return plain

def html5_extra_attributes(node, state):
    """
    @param node: the current node that could be modified
    @param state: current state
    @type state: L{Execution context<pyRdfa.state.ExecutionContext>}
    """
    def _get_literal(Pnode):
        """
        Get (recursively) the full text from a DOM Node.
    
        @param Pnode: DOM Node
        @return: string
        """
        rc = ""
        for node in Pnode.childNodes:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + node.data
            elif node.nodeType == node.ELEMENT_NODE:
                rc = rc + _get_literal(node)
        if state.options.space_preserve :
            return rc
        else :
            return re.sub(r'(\r| |\n|\t)+'," ",rc).strip()
        #return re.sub(r'(\r| |\n|\t)+',"",rc).strip()
    # end _getLiteral

    def _set_time(value) :
        if not node.hasAttribute("datatype"):
            # Check the datatype:
            dt = _format_test(value)
            if dt != plain:
                node.setAttribute("datatype",dt)
        # Finally, set the value itself
        node.setAttribute("content",value)
    # end _set_time

    if not node.hasAttribute("content") :
        # @content has top priority over the others...
        if node.hasAttribute("datetime") :
            _set_time( node.getAttribute("datetime") )
        elif node.hasAttribute("dateTime") :
            _set_time( node.getAttribute("dateTime") )
        elif node.tagName == "time" :
            # Note that a possible @datetime/@dateTime value has already been taken care of
            _set_time( _get_literal(node) )
        
def remove_rel(node, _state):
    """
    If @property and @rel/@rev are on the same element, then only CURIE and URI can appear as a rel/rev value.
    
    @param node: the current node that could be modified
    @param state: current state
    @type state: L{Execution context<pyRdfa.state.ExecutionContext>}
    """
    from ..termorcurie import termname
    def _massage_node(node,attr) :
        """The real work for remove_rel is done here, parametrized with @rel and @rev"""
        if node.hasAttribute("property") and node.hasAttribute(attr) :
            vals = node.getAttribute(attr).strip().split()
            if len(vals) != 0 :
                final_vals = [ v for v in vals if not termname.match(v) ]
                if len(final_vals) == 0 :
                    node.removeAttribute(attr)
                else :
                    node.setAttribute(attr, reduce(lambda x,y: x+' '+y,final_vals))
    
    _massage_node(node, "rev")
    _massage_node(node, "rel")




# -*- coding: utf-8 -*-
"""
Transfomer: handles the Dublin Core recommendation for XHTML for adding DC values. What this means is that:

 - DC namespaces are defined via C{<link rel="schema.XX" value="...."/>}
 - The 'XX.term' is used much like QNames in C{<link>} and C{<meta>} elements. For the latter, the namespaced names are added to a C{@property} attribute.

This transformer adds "real" namespaces and changes the DC references in link and meta elements to abide to the
RDFa namespace syntax.

@summary: Dublin Core transformer
@requires: U{RDFLib package<http://rdflib.net>}
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
@contact: Ivan Herman, ivan@w3.org
"""

"""
@version: $Id: DublinCore.py,v 1.4 2012-01-18 14:16:44 ivan Exp $
$Date: 2012-01-18 14:16:44 $
"""

def DC_transform(html, options, state):
    """
    @param html: a DOM node for the top level html element
    @param options: invocation options
    @type options: L{Options<pyRdfa.options>}
    @param state: top level execution state
    @type state: L{State<pyRdfa.state>}
    """
    from ..host import HostLanguage
    if not( options.host_language in [ HostLanguage.xhtml, HostLanguage.html5, HostLanguage.xhtml5 ] ):
        return
    
    # the head element is necessary; to be sure, the namespaces are set
    # on that level only
    head = None
    try:
        head = html.getElementsByTagName("head")[0]
    except:
        # no head....
        return

    # At first, the DC namespaces must be found
    dcprefixes = {}
    for link in html.getElementsByTagName("link"):
        if link.hasAttribute("rel"):
            rel = link.getAttribute("rel")
            uri = link.getAttribute("href")
            if uri != None and rel != None and rel.startswith("schema."):
                # bingo...
                try:
                    localname = rel.split(".")[1]
                    head.setAttributeNS("", "xmlns:"+localname,uri)
                    dcprefixes[localname] = uri
                except:
                    # problem with the split; just ignore
                    pass

    # get the link elements now to find the dc elements
    for link in html.getElementsByTagName("link"):
        if link.hasAttribute("rel"):
            newProp = ""
            for rel in link.getAttribute("rel").strip().split():
                # see if there is '.' to separate the attributes
                if rel.find(".") != -1:
                    key   = rel.split(".",1)[0]
                    lname = rel.split(".",1)[1]
                    if key in dcprefixes and lname != "":
                        # yep, this is one of those...
                        newProp += " " + key + ":" + lname
                    else:
                        newProp += " " + rel
                else:
                    newProp += " " + rel
            link.setAttribute("rel",newProp.strip())

    # do almost the same with the meta elements...
    for meta in html.getElementsByTagName("meta"):
        if meta.hasAttribute("name"):
            newProp = ""
            for name in meta.getAttribute("name").strip().split():
                # see if there is '.' to separate the attributes
                if name.find(".") != -1:
                    key   = name.split(".",1)[0]
                    lname = name.split(".",1)[1]
                    if key in dcprefixes and lname != "":
                        # yep, this is one of those...
                        newProp += " " + key + ":" + lname
                    else:
                        newProp += " " + name
                else:
                    newProp += " " + name
            meta.setAttribute("property", newProp.strip())




# -*- coding: utf-8 -*-
"""
Built-in version of the initial contexts for RDFa Core, and RDFa + HTML

@summary: Management of vocabularies, terms, and their mapping to URI-s.
@requires: U{RDFLib package<http://rdflib.net>}
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}

@var initial_context: dictionary for all the initial context data, keyed through the context URI-s
"""

"""
$Id: initialcontext.py,v 1.16 2017/10/19 12:08:11 ivan Exp $
$Date: 2017/10/19 12:08:11 $
"""

class Wrapper:
    pass

initial_context = {
    "http://www.w3.org/2011/rdfa-context/rdfa-1.1":       Wrapper(),
    "http://www.w3.org/2011/rdfa-context/xhtml-rdfa-1.1": Wrapper(),
}

initial_context["http://www.w3.org/2011/rdfa-context/rdfa-1.1"].ns = {
    'as':      'https://www.w3.org/ns/activitystreams#',
    'csvw':    'http://www.w3.org/ns/csvw#',
    'dcat':    'http://www.w3.org/ns/dcat#',
    'dqv':     'http://www.w3.org/ns/dqv#',
    'duv':     'https://www.w3.org/ns/duv#',
    'qb':      'http://purl.org/linked-data/cube#',
    'org':     'http://www.w3.org/ns/org#',
    'owl':     'http://www.w3.org/2002/07/owl#',
    'gr':      'http://purl.org/goodrelations/v1#',
    'ctag':    'http://commontag.org/ns#',
    'cc':      'http://creativecommons.org/ns#',
    'grddl':   'http://www.w3.org/2003/g/data-view#',
    'jsonld':  'http://www.w3.org/ns/json-ld#',
    'ldp':     'http://www.w3.org/ns/ldp#',
    'oa':      'http://www.w3.org/ns/oa#',
    'rif':     'http://www.w3.org/2007/rif#',
    'sioc':    'http://rdfs.org/sioc/ns#',
    'skos':    'http://www.w3.org/2004/02/skos/core#',
    'xml':     'http://www.w3.org/XML/1998/namespace',
    'rr':      'http://www.w3.org/ns/r2rml#',
    'rdfs':    'http://www.w3.org/2000/01/rdf-schema#',
    'rev':     'http://purl.org/stuff/rev#',
    'rdfa':    'http://www.w3.org/ns/rdfa#',
    'dc':      'http://purl.org/dc/terms/',
    'dcterms': 'http://purl.org/dc/terms/',
    'dc11':    'http://purl.org/dc/elements/1.1/',
    'foaf':    'http://xmlns.com/foaf/0.1/',
    'void':    'http://rdfs.org/ns/void#',
    'ical':    'http://www.w3.org/2002/12/cal/icaltzd#',
    'vcard':   'http://www.w3.org/2006/vcard/ns#',
    'wdrs':    'http://www.w3.org/2007/05/powder-s#',
    'og':      'http://ogp.me/ns#',
    'wdr':     'http://www.w3.org/2007/05/powder#',
    'rdf':     'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    'xhv':     'http://www.w3.org/1999/xhtml/vocab#',
    'xsd':     'http://www.w3.org/2001/XMLSchema#',
    'v':       'http://rdf.data-vocabulary.org/#',
    'skosxl':  'http://www.w3.org/2008/05/skos-xl#',
    'schema':  'http://schema.org/',
    'ssn':     'http://www.w3.org/ns/ssn/',
    'sosa':    'http://www.w3.org/ns/sosa/',
    'time':    'http://www.w3.org/2006/time#',
    'ma':      'http://www.w3.org/ns/ma-ont#',
    'sd':      'http://www.w3.org/ns/sparql-service-description#',
    'prov':    'http://www.w3.org/ns/prov#',
    'odrl':    'http://www.w3.org/ns/odrl/2/'
}

initial_context["http://www.w3.org/2011/rdfa-context/rdfa-1.1"].terms = {
    'describedby': 'http://www.w3.org/2007/05/powder-s#describedby',
    'role':        'http://www.w3.org/1999/xhtml/vocab#role',
    'license':     'http://www.w3.org/1999/xhtml/vocab#license',
}

initial_context["http://www.w3.org/2011/rdfa-context/rdfa-1.1"].vocabulary = ""

initial_context["http://www.w3.org/2011/rdfa-context/xhtml-rdfa-1.1"].ns = {
}

initial_context["http://www.w3.org/2011/rdfa-context/xhtml-rdfa-1.1"].vocabulary = ""

initial_context["http://www.w3.org/2011/rdfa-context/xhtml-rdfa-1.1"].terms = {
    'alternate':  'http://www.w3.org/1999/xhtml/vocab#alternate',
    'appendix':   'http://www.w3.org/1999/xhtml/vocab#appendix',
    'cite':       'http://www.w3.org/1999/xhtml/vocab#cite',
    'bookmark':   'http://www.w3.org/1999/xhtml/vocab#bookmark',
    'chapter':    'http://www.w3.org/1999/xhtml/vocab#chapter',
    'contents':   'http://www.w3.org/1999/xhtml/vocab#contents',
    'copyright':  'http://www.w3.org/1999/xhtml/vocab#copyright',
    'glossary':   'http://www.w3.org/1999/xhtml/vocab#glossary',
    'help':       'http://www.w3.org/1999/xhtml/vocab#help',
    'icon':       'http://www.w3.org/1999/xhtml/vocab#icon',
    'index':      'http://www.w3.org/1999/xhtml/vocab#index',
    'meta':       'http://www.w3.org/1999/xhtml/vocab#meta',
    'next':       'http://www.w3.org/1999/xhtml/vocab#next',
    'license':    'http://www.w3.org/1999/xhtml/vocab#license',
    'p3pv1':      'http://www.w3.org/1999/xhtml/vocab#p3pv1',
    'prev':       'http://www.w3.org/1999/xhtml/vocab#prev',
    'previous':   'http://www.w3.org/1999/xhtml/vocab#previous',
    'role':       'http://www.w3.org/1999/xhtml/vocab#role',
    'section':    'http://www.w3.org/1999/xhtml/vocab#section',
    'stylesheet': 'http://www.w3.org/1999/xhtml/vocab#stylesheet',
    'subsection': 'http://www.w3.org/1999/xhtml/vocab#subsection',
    'start':      'http://www.w3.org/1999/xhtml/vocab#start',
    'up':         'http://www.w3.org/1999/xhtml/vocab#up',
    'last':       'http://www.w3.org/1999/xhtml/vocab#last',
    'first':      'http://www.w3.org/1999/xhtml/vocab#first',
    'top':        'http://www.w3.org/1999/xhtml/vocab#top'
}



# -*- coding: utf-8 -*-
"""
Parser's execution context (a.k.a. state) object and handling. The state includes:

  - language, retrieved from C{@xml:lang} or C{@lang}
  - URI base, determined by C{<base>} or set explicitly. This is a little bit superfluous, because the current RDFa syntax does not make use of C{@xml:base}; i.e., this could be a global value.  But the structure is prepared to add C{@xml:base} easily, if needed.
  - options, in the form of an L{options<pyRdfa.options>} instance
  - a separate vocabulary/CURIE handling resource, in the form of an L{termorcurie<pyRdfa.TermOrCurie>} instance

The execution context object is also used to handle URI-s, CURIE-s, terms, etc.

@summary: RDFa parser execution context
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
"""

"""
$Id: state.py,v 1.23 2013-10-16 11:48:54 ivan Exp $
$Date: 2013-10-16 11:48:54 $
"""

from rdflib import URIRef
from rdflib import BNode

from .host import HostLanguage, accept_xml_base, accept_xml_lang, beautifying_prefixes

from .termorcurie import TermOrCurie
from . import UnresolvablePrefix, UnresolvableTerm

from . import err_URI_scheme
from . import err_illegal_safe_CURIE
from . import err_no_CURIE_in_safe_CURIE
from . import err_undefined_terms
from . import err_non_legal_CURIE_ref
from . import err_undefined_CURIE

from urllib.parse import urlparse, urlunparse, urlsplit, urljoin

class ListStructure:
    """Special class to handle the C{@inlist} type structures in RDFa 1.1; stores the "origin", i.e,
    where the list will be attached to, and the mappings as defined in the spec.
    """
    def __init__(self):
        self.mapping = {}
        self.origin      = None

#### Core Class definition
class ExecutionContext:
    """State at a specific node, including the current set of namespaces in the RDFLib sense, current language,
    the base, vocabularies, etc. The class is also used to interpret URI-s and CURIE-s to produce
    URI references for RDFLib.
    
    @ivar options: reference to the overall options
    @type options: L{Options}
    @ivar base: the 'base' URI
    @ivar parsedBase: the parsed version of base, as produced by urlparse.urlsplit
    @ivar defaultNS: default namespace (if defined via @xmlns) to be used for XML Literals
    @ivar lang: language tag (possibly None)
    @ivar term_or_curie: vocabulary management class instance
    @type term_or_curie: L{termorcurie.TermOrCurie}
    @ivar list_mapping: dictionary of arrays, containing a list of URIs key-ed via properties for lists
    @ivar node: the node to which this state belongs
    @type node: DOM node instance
    @ivar rdfa_version: RDFa version of the content
    @type rdfa_version: String
    @ivar supress_lang: in some cases, the effect of the lang attribute should be supressed for the given node, although it should be inherited down below (example: @value attribute of the data element in HTML5)
    @type supress_lang: Boolean
    @cvar _list: list of attributes that allow for lists of values and should be treated as such
    @cvar _resource_type: dictionary; mapping table from attribute name to the exact method to retrieve the URI(s). Is initialized at first instantiation.
    """

    # list of attributes that allow for lists of values and should be treated as such    
    _list = [ "rel", "rev", "property", "typeof", "role" ]
    # mapping table from attribute name to the exact method to retrieve the URI(s).
    _resource_type = {}

    def __init__(self, node, graph, inherited_state=None, base="", options=None, rdfa_version=None):
        """
        @param node: the current DOM Node
        @param graph: the RDFLib Graph
        @keyword inherited_state: the state as inherited
        from upper layers. This inherited_state is mixed with the state information
        retrieved from the current node.
        @type inherited_state: L{state.ExecutionContext}
        @keyword base: string denoting the base URI for the specific node. This overrides the possible
        base inherited from the upper layers. The 
        current XHTML+RDFa syntax does not allow the usage of C{@xml:base}, but SVG1.2 does, so this is
        necessary for SVG (and other possible XML dialects that accept C{@xml:base})
        @keyword options: invocation options, and references to warning graphs
        @type options: L{Options<pyRdfa.options>}
        """
        def remove_frag_id(uri):
            """
            The fragment ID for self.base must be removed
            """
            try:
                # To be on the safe side:-)
                t = urlparse(uri)
                return urlunparse((t[0],t[1],t[2],t[3],t[4],""))
            except:
                return uri

        # This is, conceptually, an additional class initialization, but it must be done run time, otherwise import errors show up
        if len(ExecutionContext._resource_type) == 0 :
            ExecutionContext._resource_type = {
                "href"        :    ExecutionContext._URI,
                "src"        :    ExecutionContext._URI,
                "vocab"        :   ExecutionContext._URI,

                "about"        :    ExecutionContext._CURIEorURI, 
                "resource"    :    ExecutionContext._CURIEorURI, 

                "rel"        :    ExecutionContext._TERMorCURIEorAbsURI,
                "rev"        :    ExecutionContext._TERMorCURIEorAbsURI,
                "datatype"    :    ExecutionContext._TERMorCURIEorAbsURI,
                "typeof"    :    ExecutionContext._TERMorCURIEorAbsURI,
                "property"    :    ExecutionContext._TERMorCURIEorAbsURI,
                "role"        :    ExecutionContext._TERMorCURIEorAbsURI,
            }
        #-----------------------------------------------------------------
        self.node = node
        
        #-----------------------------------------------------------------
        # Settling the base. In a generic XML, xml:base should be accepted at all levels (though this is not the
        # case in, say, XHTML...)
        # At the moment, it is invoked with a 'None' at the top level of parsing, that is
        # when the <base> element is looked for (for the HTML cases, that is)
        if inherited_state:
            self.rdfa_version = inherited_state.rdfa_version
            self.base = inherited_state.base
            self.options = inherited_state.options

            self.list_mapping = inherited_state.list_mapping
            self.new_list = False
            
            # for generic XML versions the xml:base attribute should be handled
            if self.options.host_language in accept_xml_base and node.hasAttribute("xml:base"):
                self.base = remove_frag_id(node.getAttribute("xml:base"))
        else:
            # this is the branch called from the very top            
            self.list_mapping = ListStructure()
            self.new_list = True

            if rdfa_version is not None:
                self.rdfa_version = rdfa_version
            else:
                from . import rdfa_current_version                
                self.rdfa_version = rdfa_current_version

            # This value can be overwritten by a @version attribute
            if node.hasAttribute("version"):
                top_version = node.getAttribute("version")
                if top_version.find("RDFa 1.0") != -1 or top_version.find("RDFa1.0") != -1:
                    self.rdfa_version = "1.0"
                elif top_version.find("RDFa 1.1") != -1 or top_version.find("RDFa1.1") != -1:
                    self.rdfa_version = "1.1"                        

            # this is just to play safe. I believe this should actually not happen...
            if options == None:
                from . import Options
                self.options = Options()
            else:
                self.options = options

            self.base = ""
            # handle the base element case for HTML
            if self.options.host_language in [ HostLanguage.xhtml, HostLanguage.html5, HostLanguage.xhtml5  ]:
                for bases in node.getElementsByTagName("base"):
                    if bases.hasAttribute("href"):
                        self.base = remove_frag_id(bases.getAttribute("href"))
                        continue
            elif self.options.host_language in accept_xml_base and node.hasAttribute("xml:base"):
                self.base = remove_frag_id(node.getAttribute("xml:base"))
                
            # If no local setting for base occurs, the input argument has it
            if self.base == "":
                self.base = base
                
            # Perform an extra beautification in RDFLib
            if self.options.host_language in beautifying_prefixes:
                values = beautifying_prefixes[self.options.host_language]
                for key in values:
                    graph.bind(key, values[key])

            input_info = "Input Host Language:%s, RDFa version:%s, base:%s" % (self.options.host_language, self.rdfa_version, self.base)
            self.options.add_info(input_info)

        #-----------------------------------------------------------------
        # this will be used repeatedly, better store it once and for all...        
        self.parsedBase = urlsplit(self.base)

        #-----------------------------------------------------------------
        # generate and store the local CURIE handling class instance
        self.term_or_curie = TermOrCurie(self, graph, inherited_state)

        #-----------------------------------------------------------------
        # Settling the language tags
        # @lang has priority over @xml:lang
        # it is a bit messy: the three fundamental modes (xhtml, html, or xml) are all slightly different:-(
        # first get the inherited state's language, if any
        if inherited_state:
            self.lang = inherited_state.lang
        else:
            self.lang = None
            
        self.supress_lang = False
            
            
        if self.options.host_language in [ HostLanguage.xhtml, HostLanguage.xhtml5, HostLanguage.html5 ]:
            # we may have lang and xml:lang
            if node.hasAttribute("lang"):
                lang = node.getAttribute("lang").lower()
            else:
                lang = None
            if node.hasAttribute("xml:lang"):
                xmllang = node.getAttribute("xml:lang").lower()
            else:
                xmllang = None
            # First of all, set the value, if any
            if xmllang != None:
                # this has priority
                if len(xmllang) != 0:
                    self.lang = xmllang
                else:
                    self.lang = None
            elif lang != None:
                if len(lang) != 0:
                    self.lang = lang
                else:
                    self.lang = None
            # Ideally, a warning should be generated if lang and xmllang are both present with different values. But
            # the HTML5 Parser does its magic by overriding a lang value if xmllang is present, so the potential
            # error situations are simply swallowed...
                
        elif self.options.host_language in accept_xml_lang and node.hasAttribute("xml:lang"):
                self.lang = node.getAttribute("xml:lang").lower()
                if len(self.lang) == 0:
                    self.lang = None
            
        #-----------------------------------------------------------------
        # Set the default namespace. Used when generating XML Literals
        if node.hasAttribute("xmlns"):
            self.defaultNS = node.getAttribute("xmlns")
        elif inherited_state and inherited_state.defaultNS != None:
            self.defaultNS = inherited_state.defaultNS
        else:
            self.defaultNS = None
    # end __init__

    def _URI(self, val):
        """Returns a URI for a 'pure' URI (ie, not a CURIE). The method resolves possible relative URI-s. It also
        checks whether the URI uses an unusual URI scheme (and issues a warning); this may be the result of an
        uninterpreted CURIE...
        @param val: attribute value to be interpreted
        @type val: string
        @return: an RDFLib URIRef instance
        """
        def create_URIRef(uri, check=True):
            """
            Mini helping function: it checks whether a uri is using a usual scheme before a URIRef is created. In case
            there is something unusual, a warning is generated (though the URIRef is created nevertheless)
            @param uri: (absolute) URI string
            @return: an RDFLib URIRef instance
            """
            from .    import uri_schemes
            val = uri.strip()
            if check and urlsplit(val)[0] not in uri_schemes:
                self.options.add_warning(err_URI_scheme % val.strip(), node=self.node.nodeName)
            return URIRef(val)

        def join(base, v, check=True):
            """
            Mini helping function: it makes a urljoin for the paths. Based on the python library, but
            that one has a bug: in some cases it
            swallows the '#' or '?' character at the end. This is clearly a problem with
            Semantic Web URI-s, so this is checked, too
            @param base: base URI string
            @param v: local part
            @param check: whether the URI should be checked against the list of 'existing' URI schemes
            @return: an RDFLib URIRef instance
            """
            
            joined = urljoin(base, v)
            try:
                if v[-1] != joined[-1] and (v[-1] == "#" or v[-1] == "?"):
                    return create_URIRef(joined + v[-1], check)
                else:
                    return create_URIRef(joined, check)
            except:
                return create_URIRef(joined, check)

        if val == "":
            # The fragment ID must be removed...
            return URIRef(self.base)
            
        # fall back on good old traditional URI-s.
        # To be on the safe side, let us use the Python libraries
        if self.parsedBase[0] == "":
            # base is, in fact, a local file name
            # The following call is just to be sure that some pathological cases when
            # the ':' _does_ appear in the URI but not in a scheme position is taken
            # care of properly...
            
            key = urlsplit(val)[0]
            if key == "":
                # relative URI, to be combined with local file name:
                return join(self.base, val, check = False)
            else:
                return create_URIRef(val)
        else:
            # Trust the python library...
            # Well, not quite:-) there is what is, in my view, a bug in the urljoin; in some cases it
            # swallows the '#' or '?' character at the end. This is clearly a problem with
            # Semantic Web URI-s            
            return join(self.base, val)
    # end _URI

    def _CURIEorURI(self, val):
        """Returns a URI for a (safe or not safe) CURIE. In case it is a safe CURIE but the CURIE itself
        is not defined, an error message is issued. Otherwise, if it is not a CURIE, it is taken to be a URI
        @param val: attribute value to be interpreted
        @type val: string
        @return: an RDFLib URIRef instance or None
        """
        if val == "":
            return URIRef(self.base)

        safe_curie = False
        if val[0] == '[':
            # If a safe CURIE is asked for, a pure URI is not acceptable.
            # Is checked below, and that is why the safe_curie flag is necessary
            if val[-1] != ']':
                # that is certainly forbidden: an incomplete safe CURIE
                self.options.add_warning(err_illegal_safe_CURIE % val, UnresolvablePrefix, node=self.node.nodeName)
                return None
            else:
                val = val[1:-1]
                safe_curie = True
        # There is a branch here depending on whether we are in 1.1 or 1.0 mode
        if self.rdfa_version >= "1.1":
            retval = self.term_or_curie.CURIE_to_URI(val)
            if retval == None:
                # the value could not be interpreted as a CURIE, ie, it did not produce any valid URI.
                # The rule says that then the whole value should be considered as a URI
                # except if it was part of a safe CURIE. In that case it should be ignored...
                if safe_curie:
                    self.options.add_warning(err_no_CURIE_in_safe_CURIE % val, UnresolvablePrefix, node=self.node.nodeName)
                    return None
                else:
                    return self._URI(val)
            else:
                # there is an unlikely case where the retval is actually a URIRef with a relative URI. Better filter that one out
                if isinstance(retval, BNode) == False and urlsplit(str(retval))[0] == "":
                    # yep, there is something wrong, a new URIRef has to be created:
                    return URIRef(self.base+str(retval))
                else:
                    return retval
        else:
            # in 1.0 mode a CURIE can be considered only in case of a safe CURIE
            if safe_curie:
                return self.term_or_curie.CURIE_to_URI(val)
            else:
                return self._URI(val)
    # end _CURIEorURI

    def _TERMorCURIEorAbsURI(self, val):
        """Returns a URI either for a term or for a CURIE. The value must be an NCNAME to be handled as a term; otherwise
        the method falls back on a CURIE or an absolute URI.
        @param val: attribute value to be interpreted
        @type val: string
        @return: an RDFLib URIRef instance or None
        """
        from . import uri_schemes
        # This case excludes the pure base, ie, the empty value
        if val == "":
            return None
        
        from .termorcurie import termname
        if termname.match(val):
            # This is a term, must be handled as such...            
            retval = self.term_or_curie.term_to_URI(val)
            if not retval:
                self.options.add_warning(err_undefined_terms % val, UnresolvableTerm, node=self.node.nodeName, buggy_value = val)
                return None
            else:
                return retval
        else:
            # try a CURIE
            retval = self.term_or_curie.CURIE_to_URI(val)
            if retval:
                return retval
            elif self.rdfa_version >= "1.1":
                # See if it is an absolute URI
                scheme = urlsplit(val)[0]
                if scheme == "":
                    # bug; there should be no relative URIs here
                    self.options.add_warning(err_non_legal_CURIE_ref % val, UnresolvablePrefix, node=self.node.nodeName)
                    return None
                else:
                    if scheme not in uri_schemes:
                        self.options.add_warning(err_URI_scheme % val.strip(), node=self.node.nodeName)
                    return URIRef(val)
            else:
                # rdfa 1.0 case
                self.options.add_warning(err_undefined_CURIE % val.strip(), UnresolvablePrefix, node=self.node.nodeName)
                return None
    # end _TERMorCURIEorAbsURI

    # -----------------------------------------------------------------------------------------------

    def getURI(self, attr):
        """Get the URI(s) for the attribute. The name of the attribute determines whether the value should be
        a pure URI, a CURIE, etc, and whether the return is a single element of a list of those. This is done
        using the L{ExecutionContext._resource_type} table.
        @param attr: attribute name
        @type attr: string
        @return: an RDFLib URIRef instance (or None) or a list of those
        """
        if self.node.hasAttribute(attr):
            val = self.node.getAttribute(attr)
        else:
            if attr in ExecutionContext._list:
                return []
            else:
                return None
        
        # This may raise an exception if the attr has no key. This, actually,
        # should not happen if the code is correct, but it does not harm having it here...
        try:
            func = ExecutionContext._resource_type[attr]
        except:
            # Actually, this should not happen...
            func = ExecutionContext._URI
        
        if attr in ExecutionContext._list:
            # Allows for a list
            resources = [ func(self, v.strip()) for v in val.strip().split() if v != None ]
            retval = [ r for r in resources if r != None ]
        else:
            retval = func(self, val.strip())
        return retval
    # end getURI
    
    def getResource(self, *args):
        """Get single resources from several different attributes. The first one that returns a valid URI wins.
        @param args: variable list of attribute names, or a single attribute being a list itself.
        @return: an RDFLib URIRef instance (or None):
        """
        if len(args) == 0:
            return None
        if isinstance(args[0], tuple) or isinstance(args[0], list):
            rargs = args[0]
        else:
            rargs = args
            
        for resource in rargs:
            uri = self.getURI(resource)
            if uri != None : return uri
        return None
    
    # -----------------------------------------------------------------------------------------------
    def reset_list_mapping(self, origin=None):
        """
        Reset, ie, create a new empty dictionary for the list mapping.
        """
        self.list_mapping = ListStructure()
        if origin: self.set_list_origin(origin)
        self.new_list = True

    def list_empty(self):
        """
        Checks whether the list is empty.
        @return: Boolean
        """
        return len(self.list_mapping.mapping) == 0
        
    def get_list_props(self):
        """
        Return the list of property values in the list structure
        @return: list of URIRef
        """
        return list(self.list_mapping.mapping.keys())
        
    def get_list_value(self,prop):
        """
        Return the list of values in the list structure for a specific property
        @return: list of RDF nodes
        """
        return self.list_mapping.mapping[prop]
        
    def set_list_origin(self, origin):
        """
        Set the origin of the list, ie, the subject to attach the final list(s) to
        @param origin: URIRef
        """        
        self.list_mapping.origin = origin
        
    def get_list_origin(self):
        """
        Return the origin of the list, ie, the subject to attach the final list(s) to
        @return: URIRef
        """        
        return self.list_mapping.origin
        
    def add_to_list_mapping(self, prop, resource):
        """Add a new property-resource on the list mapping structure. The latter is a dictionary of arrays;
        if the array does not exist yet, it will be created on the fly.
        
        @param prop: the property URI, used as a key in the dictionary
        @param resource: the resource to be added to the relevant array in the dictionary. Can be None; this is a dummy
        placeholder for C{<span rel="property" inlist>...</span>} constructions that may be filled in by children or siblings; if not
        an empty list has to be generated.
        """
        if prop in self.list_mapping.mapping:
            if resource != None:
                # indeed, if it is None, than it should not override anything
                if self.list_mapping.mapping[prop] == None:
                    # replacing a dummy with real content
                    self.list_mapping.mapping[prop] = [ resource ]
                else :            
                    self.list_mapping.mapping[prop].append(resource)
        else:
            if resource != None:
                self.list_mapping.mapping[prop] = [ resource ]
            else:
                self.list_mapping.mapping[prop] = None

####################



# -*- coding: utf-8 -*-
"""
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}

"""

"""
$Id: process.py,v 1.7 2012-03-23 14:06:38 ivan Exp $ $Date: 2012-03-23 14:06:38 $

"""

import sys

from rdflib import Namespace
from rdflib import RDF  as ns_rdf
from rdflib import RDFS as ns_rdfs
from rdflib import Graph

ns_owl = Namespace("http://www.w3.org/2002/07/owl#")

from ..host import MediaTypes

from ..utils import URIOpener

from . import err_outdated_cache
from . import err_unreachable_vocab
from . import err_unparsable_Turtle_vocab
from . import err_unparsable_ntriples_vocab
from . import err_unparsable_rdfa_vocab
from . import err_unrecognised_vocab_type

from .. import VocabReferenceError

from .cache import CachedVocab, xml_application_media_type
from .. import HTTPError, RDFaError

#############################################################################################################

def return_graph(uri, options, newCache = False, verify = True):
    """Parse a file, and return an RDFLib Graph. The URI's content type is checked and either one of
    RDFLib's parsers is invoked (for the Turtle, RDF/XML, and N Triple cases) or a separate RDFa processing is invoked
    on the RDFa content.

    The Accept header of the HTTP request gives a preference to Turtle, followed by RDF/XML and then HTML (RDFa), in case content negotiation is used.

    This function is used to retreive the vocabulary file and turn it into an RDFLib graph.

    @param uri: URI for the graph
    @param options: used as a place where warnings can be sent
    @param newCache: in case this is used with caching, whether a new cache is generated; that modifies the warning text
    @param verify: whether the SSL certificate should be verified
    @return: A tuple consisting of an RDFLib Graph instance and an expiration date); None if the dereferencing or the parsing was unsuccessful
    """
    def return_to_cache(msg):
        if newCache:
            options.add_warning(err_unreachable_vocab % uri, warning_type=VocabReferenceError)
        else:
            options.add_warning(err_outdated_cache % uri, warning_type=VocabReferenceError)

    retval =           None
    _expiration_date = None
    content =          None

    try:
        form = {'Accept' : 'text/html;q=0.8, application/xhtml+xml;q=0.8, text/turtle;q=1.0, application/rdf+xml;q=0.9'}
        content = URIOpener(uri, form, verify)
    except HTTPError:
        _t, value, _traceback = sys.exc_info()
        return_to_cache(value)
        return (None,None)
    except RDFaError:
        _t, value, _traceback = sys.exc_info()
        return_to_cache(value)
        return (None,None)
    except Exception:
        _t, value, _traceback = sys.exc_info()
        return_to_cache(value)
        return (None,None)

    # Store the expiration date of the newly accessed data
    expiration_date = content.expiration_date

    if content.content_type == MediaTypes.turtle:
        try:
            retval = Graph()
            retval.parse(content.data, format="n3")
        except:
            _t, value, _traceback = sys.exc_info()
            options.add_warning(err_unparsable_Turtle_vocab % (uri,value))
    elif content.content_type == MediaTypes.rdfxml:
        try:
            retval = Graph()
            retval.parse(content.data)
        except:
            _type, value, _traceback = sys.exc_info()
            options.add_warning(err_unparsable_Turtle_vocab % (uri,value))
    elif content.content_type == MediaTypes.nt:
        try:
            retval = Graph()
            retval.parse(content.data, format="nt")
        except:
            _type, value, _traceback = sys.exc_info()
            options.add_warning(err_unparsable_ntriples_vocab % (uri,value))
    elif content.content_type in [MediaTypes.xhtml, MediaTypes.html, MediaTypes.xml] or xml_application_media_type.match(content.content_type) != None:
        try:
            from .. import pyRdfa
            from .. import Options
            options = Options()
            retval = pyRdfa(options).graph_from_source(content.data)
        except:
            _type, value, _traceback = sys.exc_info()
            options.add_warning(err_unparsable_rdfa_vocab % (uri,value))
    else:
        options.add_warning(err_unrecognised_vocab_type % (uri, content.content_type))

    return (retval, expiration_date)

############################################################################################
rdftype =            ns_rdf["type"]
Property =           ns_rdf["Property"]
Class =              ns_rdfs["Class"]
subClassOf =         ns_rdfs["subClassOf"]
subPropertyOf =      ns_rdfs["subPropertyOf"]
equivalentProperty = ns_owl["equivalentProperty"]
equivalentClass =    ns_owl["equivalentClass"]

class MiniOWL:
    """
    Class implementing the simple OWL RL Reasoning required by RDFa in managing vocabulary files. This is done via
    a forward chaining process (in the L{closure} method) using a few simple rules as defined by the RDF and the OWL Semantics
    specifications.

    @ivar graph: the graph that has to be expanded
    @ivar added_triples: each cycle collects the triples that are to be added to the graph eventually.
    @type added_triples: a set, to ensure the unicity of triples being added
    """
    def __init__(self, graph, schema_semantics = False):
        self.graph = graph
        self.added_triples = None
        self.schema_semantics = schema_semantics

    def closure(self):
        """
           Generate the closure the graph. This is the real 'core'.

           The processing rules store new triples via the L{separate method<store_triple>} which stores
           them in the L{added_triples<added_triples>} array. If that array is emtpy at the end of a cycle,
           it means that the whole process can be stopped.
        """

        # Go cyclically through all rules until no change happens
        new_cycle = True
        cycle_num = 0
        while new_cycle:
            # yes, there was a change, let us go again
            cycle_num += 1

            # go through all rules, and collect the replies (to see whether any change has been done)
            # the new triples to be added are collected separately not to interfere with
            # the current graph yet
            self.added_triples = set()

            # Execute all the rules; these might fill up the added triples array
            for t in self.graph: self.rules(t)

            # Add the tuples to the graph (if necessary, that is). If any new triple has been generated, a new cycle
            # will be necessary...
            new_cycle = len(self.added_triples) > 0

            for t in self.added_triples: self.graph.add(t)

    def store_triple(self, t):
        """
        In contrast to its name, this does not yet add anything to the graph itself, it just stores the tuple in an
        L{internal set<added_triples>}. (It is important for this to be a set: some of the rules in the various closures may
        generate the same tuples several times.) Before adding the tuple to the set, the method checks whether
        the tuple is in the final graph already (if yes, it is not added to the set).

        The set itself is emptied at the start of every processing cycle; the triples are then effectively added to the
        graph at the end of such a cycle. If the set is
        actually empty at that point, this means that the cycle has not added any new triple, and the full processing can stop.

        @param t: the triple to be added to the graph, unless it is already there
        @type t: a 3-element tuple of (s,p,o)
        """
        if t not in self.graph:
            self.added_triples.add(t)

    def rules(self, t):
        """
            Go through the OWL-RL entailement rules prp-spo1, prp-eqp1, prp-eqp2, cax-sco, cax-eqc1, and cax-eqc2 by extending the graph.
            @param t: a triple (in the form of a tuple)
        """
        s, p, o = t
        if self.schema_semantics:
            # extra resonings on the vocabulary only to reduce the overall load by reducing the expected number of chaining cycles
            if p == subPropertyOf:
                for _z, _y, xxx in self.graph.triples((o, subPropertyOf, None)):
                    self.store_triple((s, subPropertyOf, xxx))
            elif p == equivalentProperty:
                for _z, _y, xxx in self.graph.triples((o, equivalentProperty, None)):
                    self.store_triple((s, equivalentProperty, xxx))
                for xxx, _y, _z in self.graph.triples((None, equivalentProperty, s)):
                    self.store_triple((xxx, equivalentProperty, o))
            elif p == subClassOf:
                for _z, _y, xxx in self.graph.triples((o, subClassOf, None)):
                    self.store_triple((s, subClassOf, xxx))
            elif p == equivalentClass:
                for _z, _y, xxx in self.graph.triples((o, equivalentClass, None)):
                    self.store_triple((s, equivalentClass, xxx))
                for xxx, _y, _z in self.graph.triples((None, equivalentClass, s)):
                    self.store_triple((xxx, equivalentClass, o))
        else:
            if p == subPropertyOf:
                # prp-spo1
                for zzz, _z, www in self.graph.triples((None, s, None)):
                    self.store_triple((zzz, o, www))
            elif p == equivalentProperty:
                # prp-eqp1
                for zzz, _z, www in self.graph.triples((None, s, None)):
                    self.store_triple((zzz, o, www))
                # prp-eqp2
                for zzz, _z, www in self.graph.triples((None, o, None)):
                    self.store_triple((zzz, s, www))
            elif p == subClassOf:
                # cax-sco
                for vvv, _y, _z in self.graph.triples((None, rdftype, s)):
                    self.store_triple((vvv, rdftype, o))
            elif p == equivalentClass:
                # cax-eqc1
                for vvv, _y, _z in self.graph.triples((None, rdftype, s)):
                    self.store_triple((vvv, rdftype, o))
                # cax-eqc2
                for vvv, _y, _z in self.graph.triples((None, rdftype, o)):
                    self.store_triple((vvv, rdftype, s))

########################################################################################################

def process_rdfa_sem(graph, options):
    """
    Expand the graph through the minimal RDFS and OWL rules defined for RDFa.

    The expansion is done in several steps:
     1. the vocabularies are retrieved from the incoming graph (there are RDFa triples generated for that)
     2. all vocabularies are merged into a separate vocabulary graph
     3. the RDFS/OWL expansion is done on the vocabulary graph, to take care of all the subproperty, subclass, etc, chains
     4. the (expanded) vocabulary graph content is added to the incoming graph
     5. the incoming graph is expanded
     6. the triples appearing in the vocabulary graph are removed from the incoming graph, to avoid unnecessary extra triples from the data

    @param graph: an RDFLib Graph instance, to be expanded
    @param options: options as defined for the RDFa run; used to generate warnings
    @type options: L{pyRdfa.Options}
    """
    # 1. collect the vocab URI-s
    vocabs = set()
    from ...pyRdfa import RDFA_VOCAB
    for ((_s, _p, v)) in graph.triples((None, RDFA_VOCAB, None)):
        vocabs.add((str(v)))

    if len(vocabs) >= 0:
        # 2. get all the vocab graphs
        vocab_graph = Graph()
        for uri in vocabs:
            if options.vocab_cache:
                v_graph = CachedVocab(uri, options).graph
            else:
                v_graph, _exp_date = return_graph(uri, options)
            if v_graph != None:
                for t in v_graph:
                    vocab_graph.add(t)

        # 3. Get the closure of the vocab graph; this will take care of local subproperty, etc, statements
        # Strictly speaking this is not necessary, but will speed up processing, because it may save chaining cycles on the
        # real graph
        MiniOWL(vocab_graph, schema_semantics = True).closure()

        # 4. Now get the vocab graph content added to the default graph
        for t in vocab_graph:
            graph.add(t)

        # 5. get the graph expanded through RDFS
        MiniOWL(graph).closure()

        # 4. clean up the graph by removing the schema triples
        for t in vocab_graph : graph.remove(t)

    # That was it...
    return graph




# -*- coding: utf-8 -*-
"""
Management of vocabularies, terms, and their mapping to URI-s. The main class of this module (L{TermOrCurie}) is,
conceptually, part of the overall state of processing at a node (L{state.ExecutionContext}) but putting it into a separate
module makes it easider to maintain.

@summary: Management of vocabularies, terms, and their mapping to URI-s.
@requires: U{RDFLib package<http://rdflib.net>}
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}

@var XHTML_PREFIX: prefix for the XHTML vocabulary URI (set to 'xhv')
@var XHTML_URI: URI prefix of the XHTML vocabulary
@var ncname: Regular expression object for NCNAME
@var termname: Regular expression object for a term
@var xml_application_media_type: Regular expression object for a general XML application media type
"""

"""
$Id: termorcurie.py,v 1.12 2013-10-16 11:48:54 ivan Exp $
$Date: 2013-10-16 11:48:54 $
"""

import re

from urllib.parse import urlsplit


from rdflib import URIRef
from rdflib import BNode
from rdflib import Namespace

from .utils import quote_URI
from .host import predefined_1_0_rel, warn_xmlns_usage
from . import IncorrectPrefixDefinition, RDFA_VOCAB, UnresolvableReference, PrefixRedefinitionWarning

from . import err_redefining_URI_as_prefix
from . import err_xmlns_deprecated
from . import err_bnode_local_prefix
from . import err_col_local_prefix
from . import err_missing_URI_prefix
from . import err_invalid_prefix
from . import err_no_default_prefix
from . import err_prefix_and_xmlns
from . import err_non_ncname_prefix
from . import err_absolute_reference
from . import err_query_reference
from . import err_fragment_reference
from . import err_prefix_redefinition


# Regular expression object for NCNAME
ncname   = re.compile("^[A-Za-z][A-Za-z0-9._-]*$")

# Regular expression object for term name
termname = re.compile("^[A-Za-z]([A-Za-z0-9._-]|/)*$")

# Regular expression object for a general XML application media type
xml_application_media_type = re.compile(r"application/[a-zA-Z0-9]+\+xml")

XHTML_PREFIX = "xhv"
XHTML_URI    = "http://www.w3.org/1999/xhtml/vocab#"

#### Managing blank nodes for CURIE-s: mapping from local names to blank nodes.
_bnodes = {}
_empty_bnode = BNode()

####

class InitialContext:
    """
    Get the initial context values. In most cases this class has an empty content, except for the
    top level (in case of RDFa 1.1). Each L{TermOrCurie} class has one instance of this class. It provides initial
    mappings for terms, namespace prefixes, etc, that the top level L{TermOrCurie} instance uses for its own initialization.
    
    @ivar terms: collection of all term mappings
    @type terms: dictionary
    @ivar ns: namespace mapping
    @type ns: dictionary
    @ivar vocabulary: default vocabulary
    @type vocabulary: string
    """    
    
    def __init__(self, state, top_level):
        """
        @param state: the state behind this term mapping
        @type state: L{state.ExecutionContext}
        @param top_level : whether this is the top node of the DOM tree (the only place where initial contexts are handled)
        @type top_level : boolean
        """        
        self.state = state

        # This is to store the local terms
        self.terms  = {}
        # This is to store the local Namespaces (a.k.a. prefixes)
        self.ns     = {}
        # Default vocabulary
        self.vocabulary = None
        
        if state.rdfa_version < "1.1" or top_level == False:
            return
        
        from .initialcontext import initial_context as context_data
        from .host import initial_contexts as context_ids
        from .host import default_vocabulary

        for i in context_ids[state.options.host_language]:
            # This gives the id of a initial context, valid for this media type:
            data = context_data[i]
            
            # Merge the context data with the overall definition
            if state.options.host_language in default_vocabulary:
                self.vocabulary = default_vocabulary[state.options.host_language]
            elif data.vocabulary != "":
                self.vocabulary = data.vocabulary

            for key in data.terms:
                self.terms[key] = URIRef(data.terms[key])
            for key in data.ns:
                self.ns[key] = (Namespace(data.ns[key]),False)


##################################################################################################################

class TermOrCurie:
    """
    Wrapper around vocabulary management, ie, mapping a term to a URI, as well as a CURIE to a URI. Each instance of this class belongs to a
    "state", instance of L{state.ExecutionContext}. Context definitions are managed at initialization time.
    
    (In fact, this class is, conceptually, part of the overall state at a node, and has been separated here for an
    easier maintenance.)
    
    The class takes care of the stack-like behavior of vocabulary items, ie, inheriting everything that is possible
    from the "parent". At initialization time, this works through the prefix definitions (i.e., C{@prefix} or C{@xmln:} attributes)
    and/or C{@vocab} attributes.
    
    @ivar state: State to which this instance belongs
    @type state: L{state.ExecutionContext}
    @ivar graph: The RDF Graph under generation
    @type graph: rdflib.Graph
    @ivar terms: mapping from terms to URI-s
    @type terms: dictionary
    @ivar ns: namespace declarations, ie, mapping from prefixes to URIs
    @type ns: dictionary
    @ivar default_curie_uri: URI for a default CURIE
    """
    def __init__(self, state, graph, inherited_state):
        """Initialize the vocab bound to a specific state. 
        @param state: the state to which this vocab instance belongs to
        @type state: L{state.ExecutionContext}
        @param graph: the RDF graph being worked on
        @type graph: rdflib.Graph
        @param inherited_state: the state inherited by the current state. 'None' if this is the top level state.
        @type inherited_state: L{state.ExecutionContext}
        """
        def check_prefix(pr):
            from . import uri_schemes
            if pr in uri_schemes:
                # The prefix being defined is a registered URI scheme, better avoid it...
                state.options.add_warning(err_redefining_URI_as_prefix % pr, node=state.node.nodeName)
                
        self.state    = state
        self.graph    = graph
        
        # --------------------------------------------------------------------------------
        # This is set to non-void only on the top level and in the case of 1.1
        default_vocab = InitialContext(self.state, inherited_state == None)
        
        # Set the default CURIE URI
        if inherited_state == None:
            # This is the top level...
            self.default_curie_uri = Namespace(XHTML_URI)
            # self.graph.bind(XHTML_PREFIX, self.default_curie_uri)
        else:
            self.default_curie_uri = inherited_state.term_or_curie.default_curie_uri
        
        # --------------------------------------------------------------------------------
        # Set the default term URI
        # This is a 1.1 feature, ie, should be ignored if the version is < 1.0
        if state.rdfa_version >= "1.1":
            # that is the absolute default setup...
            if inherited_state == None:
                self.default_term_uri = None
            else:
                self.default_term_uri = inherited_state.term_or_curie.default_term_uri
                
            # see if the initial context has defined a default vocabulary:
            if default_vocab.vocabulary:
                self.default_term_uri = default_vocab.vocabulary
                
            # see if there is local vocab that would override previous settings
            # However, care should be taken with the vocab="" value that should not become a URI...
            # Indeed, this value is used to 'vipe out', ie, get back to the default vocabulary...
            if self.state.node.hasAttribute("vocab") and self.state.node.getAttribute("vocab") == "":
                self.default_term_uri = default_vocab.vocabulary
            else:
                def_term_uri = self.state.getURI("vocab")
                if def_term_uri and def_term_uri != "" :            
                    self.default_term_uri = def_term_uri
                    self.graph.add((URIRef(self.state.base),RDFA_VOCAB,URIRef(def_term_uri)))
        else:
            self.default_term_uri = None
        
        # --------------------------------------------------------------------------------
        # The simpler case: terms, adding those that have been defined by a possible initial context
        if inherited_state is None:
            # this is the vocabulary belonging to the top level of the tree!
            self.terms = {}
            if state.rdfa_version >= "1.1":
                # Simply get the terms defined by the default vocabularies. There is no need for merging
                for key in default_vocab.terms:
                    self.terms[key] = default_vocab.terms[key]
            else:
                # The terms are hardwired...
                for key in predefined_1_0_rel:
                    self.terms[key] = URIRef(XHTML_URI + key)
        else:
            # just refer to the inherited terms
            self.terms = inherited_state.term_or_curie.terms

        #-----------------------------------------------------------------
        # the locally defined namespaces
        ns_dict = {}
        # locally defined xmlns namespaces, necessary for correct XML Literal generation
        xmlns_dict = {}

        # Add the locally defined namespaces using the xmlns: syntax
        for i in range(0, state.node.attributes.length):
            attr = state.node.attributes.item(i)
            if attr.name.find('xmlns:') == 0 :    
                # yep, there is a namespace setting
                prefix = attr.localName
                if prefix != "" : # exclude the top level xmlns setting...
                    if state.rdfa_version >= "1.1" and state.options.host_language in warn_xmlns_usage:
                        state.options.add_warning(err_xmlns_deprecated % prefix, IncorrectPrefixDefinition, node=state.node.nodeName)
                    if prefix == "_":
                        state.options.add_warning(err_bnode_local_prefix, IncorrectPrefixDefinition, node=state.node.nodeName)
                    elif prefix.find(':') != -1:
                        state.options.add_warning(err_col_local_prefix % prefix, IncorrectPrefixDefinition, node=state.node.nodeName)
                    else :                    
                        # quote the URI, ie, convert special characters into %.. This is
                        # true, for example, for spaces
                        uri = quote_URI(attr.value, state.options)
                        # create a new RDFLib Namespace entry
                        ns = Namespace(uri)
                        # Add an entry to the dictionary if not already there (priority is left to right!)
                        if state.rdfa_version >= "1.1":
                            pr = prefix.lower()
                        else:
                            pr = prefix
                        ns_dict[pr]       = ns
                        xmlns_dict[pr] = ns
                        self.graph.bind(pr,ns)
                        check_prefix(pr)

        # Add the locally defined namespaces using the @prefix syntax
        # this may override the definition @xmlns
        if state.rdfa_version >= "1.1" and state.node.hasAttribute("prefix"):
            pr = state.node.getAttribute("prefix")
            if pr != None:
                # separator character is whitespace
                pr_list = pr.strip().split()
                # range(0, len(pr_list), 2) 
                for i in range(len(pr_list) - 2, -1, -2):
                    prefix = pr_list[i]
                    # see if there is a URI at all
                    if i == len(pr_list) - 1:
                        state.options.add_warning(err_missing_URI_prefix % (prefix,pr), node=state.node.nodeName)
                        break
                    else:
                        value = pr_list[i+1]
                    
                    # see if the value of prefix is o.k., ie, there is a ':' at the end
                    if prefix[-1] != ':':
                        state.options.add_warning(err_invalid_prefix % (prefix,pr), IncorrectPrefixDefinition, node=state.node.nodeName)
                        continue
                    elif prefix == ":":
                        state.options.add_warning(err_no_default_prefix % pr, IncorrectPrefixDefinition, node=state.node.nodeName)
                        continue                        
                    else:
                        prefix = prefix[:-1]
                        uri    = Namespace(quote_URI(value, state.options))
                        if prefix == "":
                            #something to be done here
                            self.default_curie_uri = uri
                        elif prefix == "_":
                            state.options.add_warning(err_bnode_local_prefix, IncorrectPrefixDefinition, node=state.node.nodeName)
                        else:
                            # last check: is the prefix an NCNAME?
                            if ncname.match(prefix):
                                real_prefix = prefix.lower()
                                ns_dict[real_prefix] = uri
                                self.graph.bind(real_prefix,uri)
                                # Additional warning: is this prefix overriding an existing xmlns statement with a different URI? if
                                # so, that may lead to discrepancies between an RDFa 1.0 and RDFa 1.1 run...
                                if (prefix in xmlns_dict and xmlns_dict[prefix] != uri) or (real_prefix in xmlns_dict and xmlns_dict[real_prefix] != uri):
                                    state.options.add_warning(err_prefix_and_xmlns % (real_prefix,real_prefix), node=state.node.nodeName)
                                check_prefix(real_prefix)

                            else:
                                state.options.add_warning(err_non_ncname_prefix % (prefix,pr), IncorrectPrefixDefinition, node=state.node.nodeName)

        # See if anything has been collected at all.
        # If not, the namespaces of the incoming state is
        # taken over by reference. Otherwise that is copied to the
        # the local dictionary
        if inherited_state == None:
            self.default_prefixes = default_vocab.ns
            inherited_prefixes    = {}
        else:
            self.default_prefixes = inherited_state.term_or_curie.default_prefixes
            inherited_prefixes    = inherited_state.term_or_curie.ns

        if len(ns_dict) == 0:
            self.ns = inherited_prefixes
        else:
            self.ns = {}
            for key in inherited_prefixes : self.ns[key] = inherited_prefixes[key]
            for key in ns_dict : 
                if (key in inherited_prefixes and ns_dict[key] != inherited_prefixes[key]) or (key in self.default_prefixes and ns_dict[key] != self.default_prefixes[key][0]):
                    state.options.add_warning(err_prefix_redefinition % key, PrefixRedefinitionWarning, node=state.node.nodeName)
                self.ns[key] = ns_dict[key]

        
        # the xmlns prefixes have to be stored separately, again for XML Literal generation    
        self.xmlns = {}
        if len(xmlns_dict) == 0 and inherited_state:
            self.xmlns = inherited_state.term_or_curie.xmlns
        else:
            if inherited_state:
                for key in inherited_state.term_or_curie.xmlns    : self.xmlns[key] = inherited_state.term_or_curie.xmlns[key]
                for key in xmlns_dict                            : self.xmlns[key] = xmlns_dict[key]
            else:
                self.xmlns = xmlns_dict
    # end __init__

    def _check_reference(self, val):
        """Checking the CURIE reference for correctness. It is probably not 100% foolproof, but may take care
        of some of the possible errors. See the URI RFC for the details.
        """
        def char_check(s, not_allowed = ['#','[',']']):
            for c in not_allowed:
                if s.find(c) != -1 : return False
            return True
        # Creating an artificial http URI to fool the urlparse module...
        _scheme, netloc, _url, query, fragment = urlsplit('http:' + val)
        if netloc != "" and self.state.rdfa_version >= "1.1":
            self.state.options.add_warning(err_absolute_reference % (netloc, val), UnresolvableReference, node=self.state.node.nodeName)
            return False
        elif not char_check(query):
            self.state.options.add_warning(err_query_reference % (query, val), UnresolvableReference, node=self.state.node.nodeName)
            return False
        elif not char_check(fragment):
            self.state.options.add_warning(err_fragment_reference % (fragment, val), UnresolvableReference, node=self.state.node.nodeName)
            return False
        else:
            return True

    def CURIE_to_URI(self, val):
        """CURIE to URI mapping. 
        
        This method does I{not} take care of the last step of CURIE processing, ie, the fact that if
        it does not have a CURIE then the value is used a URI. This is done on the caller's side, because this has
        to be combined with base, for example. The method I{does} take care of BNode processing, though, ie,
        CURIE-s of the form "_:XXX".
        
        @param val: the full CURIE
        @type val: string
        @return: URIRef of a URI or None.
        """
        # Just to be on the safe side:
        if val == "":
            return None
        elif val == ":":
            if self.default_curie_uri:
                return URIRef(self.default_curie_uri)
            else:
                return None
        
        # See if this is indeed a valid CURIE, ie, it can be split by a colon
        curie_split = val.split(':',1)
        if len(curie_split) == 1:
            # there is no ':' character in the string, ie, it is not a valid CURIE
            return None
        else:
            if self.state.rdfa_version >= "1.1":
                prefix    = curie_split[0].lower()
            else:
                prefix    = curie_split[0]
            reference = curie_split[1]

            #if len(reference) > 0:
            #    if self.state.rdfa_version >= "1.1" and (len(prefix) == 0 or prefix in self.ns) and reference.startswith('//'):
            #        # This has been defined as illegal in RDFa 1.1
            #        self.state.options.add_warning(err_absolute_reference % (reference, val), UnresolvableReference, node=self.state.node.nodeName)
            #        return None
            #    if reference[0] == ":":
            #        return None
            
            # first possibility: empty prefix
            if len(prefix) == 0:
                if self.default_curie_uri and self._check_reference(reference):
                    return self.default_curie_uri[reference]
                else:
                    return None
            else:
                # prefix is non-empty; can be a bnode
                if prefix == "_":
                    # yep, BNode processing. There is a difference whether the reference is empty or not...
                    if len(reference) == 0:
                        return _empty_bnode
                    else:
                        # see if this variable has been used before for a BNode
                        if reference in _bnodes:
                            return _bnodes[reference]
                        else:
                            # a new bnode...
                            retval = BNode()
                            _bnodes[reference] = retval
                            return retval
                # check if the prefix is a valid NCNAME
                elif ncname.match(prefix):
                    # see if there is a binding for this:                    
                    if prefix in self.ns and self._check_reference(reference):
                        # yep, a binding has been defined!
                        if len(reference) == 0:
                            return URIRef(str(self.ns[prefix]))
                        else:
                            return self.ns[prefix][reference]
                    elif prefix in self.default_prefixes and self._check_reference(reference):
                        # this has been defined through the default context
                        if len(reference) == 0:
                            return URIRef(str(self.default_prefixes[prefix][0]))
                        else:
                            (ns,used) = self.default_prefixes[prefix]
                            # lazy binding of prefixes (to avoid unnecessary prefix definitions in the serializations at the end...)
                            if not used:
                                self.graph.bind(prefix,ns)
                                self.default_prefixes[prefix] = (ns,True)
                            return ns[reference]
                    else:
                        # no definition for this thing...
                        return None
                else:
                    return None
    # end CURIE_to_URI

    def term_to_URI(self, term):
        """A term to URI mapping, where term is a simple string and the corresponding
        URI is defined via the @vocab (ie, default term uri) mechanism. Returns None if term is not defined
        @param term: string
        @return: an RDFLib URIRef instance (or None)
        """
        if len(term) == 0 : return None

        if termname.match(term):
            # It is a valid NCNAME
            
            # First of all, a @vocab nukes everything. That has to be done first...
            if self.default_term_uri != None:
                return URIRef(self.default_term_uri + term)

            # For default terms, the algorithm is (see 7.4.3 of the document): first make a case sensitive match;
            # if that fails than make a case insensive one            
            # 1. simple, case sensitive test:
            if term in self.terms:
                # yep, term is a valid key as is
                # lazy binding of the xhv prefix for terms...
                self.graph.bind(XHTML_PREFIX, XHTML_URI)
                return self.terms[term]
                
            # 2. case insensitive test
            for defined_term in self.terms:
                if term.lower() == defined_term.lower():
                    # lazy binding of the xhv prefix for terms...
                    self.graph.bind(XHTML_PREFIX, XHTML_URI)
                    return self.terms[defined_term]

        # If it got here, it is all wrong...
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Run the pyRdfa package locally, ie, on a local file

"""
# You may want to adapt this to your environment...
import sys, getopt, platform, logging

sys.path.insert(0,"/Users/ivan/Library/Python")
sys.path.insert(0,"/Users/ivan/Library/Python/RDFa")

from pyRdfa                        import pyRdfa
from pyRdfa.transform.metaname     import meta_transform
from pyRdfa.transform.OpenID       import OpenID_transform
from pyRdfa.transform.DublinCore   import DC_transform
from pyRdfa.options                import Options

extraTransformers = [
	# containers_collections,
	#OpenID_transform,
	#DC_transform,
	# meta_transform
]

###########################################

usageText="""Usage: %s -[vjxtnpzsb:g:ryle] [filename[s]]
where:
  -x: output format RDF/XML
  -t: output format Turtle (default)
  -j: output format JSON-LD
  -n: output format N Triples
  -p: output format pretty RDF/XML
  -z: exceptions should be returned as graphs instead of exceptions raised
  -b: give the base URI; if a file name is given, this can be left empty and the file name is used
  -s: whitespace on plain literals are not preserved (default: preserved, per RDFa syntax document); this is a non-standard feture
  -l: run in RDFa 1.1 Lite mode (non-RDFa Lite generate warnings) (default: False)
  -r: report on the details of the vocabulary caching process
  -y: bypass the vocabulary cache checking, generate a new cache every time (good for debugging) (default:False)
  -v: perform vocabulary expansion (default: False)
  -g: value can be 'output', 'processor', 'output,processor' or 'processor,output'; controls which graphs are returned
  -e: embedded (in a <script> element) Turtle content or RDF/XML (in its own namespace) _not_ parsed and added to the output graph (default:True, i.e., parsed)
  -w: value is a filename, to be used instead of the default standard error for logging errors, warnings, etc.

'Filename' can be a local file name or a URI. In case there is no filename, stdin is used.

The -g option may be unnecessary, the script tries to make a guess based on a default xmlns value for XHTML or SVG.
"""

def usage() :
	print(usageText % sys.argv[0])

format                 = "turtle"
extras                 = []
value                  = ""
space_preserve         = True
base                   = ""
value                  = []
rdfOutput              = False
output_default_graph   = True
output_processor_graph = True
vocab_cache_report     = False
refresh_vocab_cache    = False
vocab_expansion        = False
vocab_cache            = True
embedded_rdf           = True
check_lite             = False
log					   = None

try :
	opts, value = getopt.getopt(sys.argv[1:],"vxetjnpzsb:g:rylw:",['graph='])
	for o,a in opts:
		if o == "-t" :
			format = "turtle"
		elif o == "-n" :
			format = "nt"
		elif o == "-j" :
			format = "json-ld"
		elif o == "-p" or o == "-x":
			format = "pretty-xml"
		elif o == "-z" :
			rdfOutput = True
		elif o == "-b" :
			base = a
		elif o == "-e" :
			embedded_rdf = False
		elif o == "-s" :
			space_preserve = False
		elif o == "-l" :
			check_lite = True
		elif o == "-r" :
			vocab_cache_report = True
		elif o == "-v" :
			vocab_expansion = True
		elif o == "-y" :
			bypass_vocab_cache = True
		elif o in ("-g", "--graph") :
			if a == "processor" :
				output_default_graph 	= False
				output_processor_graph 	= True
			elif a == "processor,default" or a == "default,processor" :
				output_processor_graph 	= True
			elif a == "default" :
				output_default_graph 	= True
				output_processor_graph 	= False
		elif o == "-w" :
			log = a
		else :
			usage()
			sys.exit(1)
except :
	usage()
	sys.exit(1)

if log is not None :
	logging.basicConfig(filename=log)
else :
	logging.basicConfig()

options = Options(output_default_graph = output_default_graph,
				  output_processor_graph = output_processor_graph,
				  space_preserve=space_preserve,
				  transformers = extras,
				  embedded_rdf = embedded_rdf,
				  vocab_expansion = vocab_expansion,
				  vocab_cache = vocab_cache,
				  vocab_cache_report = vocab_cache_report,
				  refresh_vocab_cache = refresh_vocab_cache,
				  check_lite = check_lite,
				  experimental_features = True
)

processor = pyRdfa(options, base)
if len(value) >= 1 :
	print(processor.rdf_from_sources(value, outputFormat = format, rdfOutput = rdfOutput))
else :
	print(processor.rdf_from_source(sys.stdin, outputFormat = format, rdfOutput = rdfOutput))





# -*- coding: utf-8 -*-
"""
Simple transfomer: handle OpenID elements. Ie: an openid namespace is added and the usual
'link' elements for openid are exchanged against a namespaced version.

@summary: OpenID transformer module.
@requires: U{RDFLib package<http://rdflib.net>}
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
@contact: Ivan Herman, ivan@w3.org
@var OPENID_NS: the OpenID URI used in the package
"""

"""
$Id: OpenID.py,v 1.4 2012-01-18 14:16:44 ivan Exp $
$Date: 2012-01-18 14:16:44 $
"""

OPENID_NS = "http://xmlns.openid.net/auth#"


def OpenID_transform(html, options, _state):
    """
    Replace C{openid.XXX} type C{@rel} attribute values in C{<link>} elements by C{openid:XXX}. The openid URI is also
    added to the top level namespaces with the C{openid:} local name.

    @param html: a DOM node for the top level html element
    @param options: invocation options
    @type options: L{Options<pyRdfa.options>}
    @param state: top level execution state
    @type state: L{State<pyRdfa.state>}
    """
    from ..host import HostLanguage
    if not(options.host_language in [HostLanguage.xhtml, HostLanguage.html5, HostLanguage.xhtml5]):
        return

    # the head element is necessary; to be sure, the namespaces are set
    # on that level only
    head = None
    try:
        head = html.getElementsByTagName("head")[0]
    except:
        # no head....
        return

    foundOpenId = False
    for link in html.getElementsByTagName("link"):
        if link.hasAttribute("rel"):
            rel = link.getAttribute("rel")
            newProp = ""
            for n in rel.strip().split():
                if n.startswith("openid."):
                    newProp += " " + n.replace("openid.", "openid:")
                    foundOpenId = True
                else:
                    newProp += " " + n
            link.setAttribute("rel",newProp.strip())

    # Add the OpenId namespace if necessary
    if foundOpenId and not head.hasAttribute("xmlns:openid"):
        head.setAttributeNS("", "xmlns:openid", OPENID_NS)




# -*- coding: utf-8 -*-
"""
Various utilities for pyRdfa.

Most of the utilities are straightforward.

@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}


"""

"""
$Id: utils.py,v 1.18 2016/12/08 10:13:34 ivan Exp $
$Date: 2016/12/08 10:13:34 $
"""
import sys, datetime

from urllib.parse import urljoin, quote
from http.server import BaseHTTPRequestHandler
from urllib.error import HTTPError as urllib_HTTPError

from .extras.httpheader import content_type, parse_http_datetime


from .host import preferred_suffixes

#########################################################################################################
# Handling URIs
class URIOpener:
    """A wrapper around the urllib2 method to open a resource. Beyond accessing the data itself, the class
    sets a number of instance variable that might be relevant for processing.
    The class also adds an accept header to the outgoing request, namely
    text/html and application/xhtml+xml (unless set explicitly by the caller).
    
    If the content type is set by the server, the relevant HTTP response field is used. Otherwise,
    common suffixes are used (see L{host.preferred_suffixes}) to set the content type (this is really of importance
    for C{file:///} URI-s). If none of these works, the content type is empty.
        
    Interpretation of the content type for the return is done by Deron Meranda's U{httpheader module<http://deron.meranda.us/>}.
    
    @ivar data: the real data, ie, a file-like object
    @ivar headers: the return headers as sent back by the server
    @ivar content_type: the content type of the resource or the empty string, if the content type cannot be determined
    @ivar location: the real location of the data (ie, after possible redirection and content negotiation)
    @ivar last_modified_date: sets the last modified date if set in the header, None otherwise
    @ivar expiration_date: sets the expiration date if set in the header, I{current UTC plus one day} otherwise (this is used for caching purposes, hence this artificial setting)
    """
    CONTENT_LOCATION = 'Content-Location'
    CONTENT_TYPE =     'Content-Type'
    LAST_MODIFIED =    'Last-Modified'
    EXPIRES =          'Expires'

    def __init__(self, name, additional_headers={}, verify=True):
        """
        @param name: URL to be opened
        @keyword additional_headers: additional HTTP request headers to be added to the call
        """        
        try:
            # Note the removal of the fragment ID. This is necessary, per the HTTP spec
            url = name.split('#')[0]
            if 'Accept' not in additional_headers:
                additional_headers['Accept'] = 'text/html, application/xhtml+xml'
                
            import requests
            # For security reason certificate verification is now done by default. But, can be
            # disabled for sites still go wrong because the cerficates are not o.k. with request...
            r = requests.get(url, headers=additional_headers, verify=verify)
            self.data = r.content
            self.headers = r.headers
            
            if URIOpener.CONTENT_TYPE in self.headers:
                # The call below will remove the possible media type parameters, like charset settings
                ct = content_type(self.headers[URIOpener.CONTENT_TYPE])
                self.content_type = ct.media_type
                if 'charset' in ct.parmdict:
                    self.charset = ct.parmdict['charset']
                else:
                    self.charset = None
                # print
            else:
                # check if the suffix can be used for the content type; this may be important
                # for file:// type URI or if the server is not properly set up to return the right
                # mime type
                self.charset = None
                self.content_type = ""
                for suffix in preferred_suffixes.keys():
                    if name.endswith(suffix):
                        self.content_type = preferred_suffixes[suffix]
                        break
            
            if URIOpener.CONTENT_LOCATION in self.headers:
                self.location = urljoin(r.url, self.headers[URIOpener.CONTENT_LOCATION])
            else:
                self.location = name
            
            self.expiration_date = datetime.datetime.utcnow() + datetime.timedelta(days=1)
            if URIOpener.EXPIRES in self.headers:
                try:
                    # Thanks to Deron Meranda for the HTTP date conversion method...
                    self.expiration_date = parse_http_datetime(self.headers[URIOpener.EXPIRES])
                except:
                    # The Expires date format was wrong, sorry, forget it...
                    pass

            self.last_modified_date = None
            if URIOpener.LAST_MODIFIED in self.headers:
                try:
                    # Thanks to Deron Meranda for the HTTP date conversion method...
                    self.last_modified_date = parse_http_datetime(self.headers[URIOpener.LAST_MODIFIED])
                except:
                    # The last modified date format was wrong, sorry, forget it...
                    pass
                
        except urllib_HTTPError:
            e = sys.exc_info()[1]
            from . import HTTPError
            msg = BaseHTTPRequestHandler.responses[e.code]
            raise HTTPError('%s' % msg[1], e.code)
        except Exception:
            e = sys.exc_info()[1]
            from . import RDFaError
            raise RDFaError('%s' % e)

#########################################################################################################

# 'safe' characters for the URI quoting, ie, characters that can safely stay as they are. Other 
# special characters are converted to their %.. equivalents for namespace prefixes
_unquotedChars = r':/\?=#~'
_warnChars = [' ','\n','\r','\t']

def quote_URI(uri, options=None):
    """
    'quote' a URI, ie, exchange special characters for their '%..' equivalents. Some of the characters
    may stay as they are (listed in L{_unquotedChars}. If one of the characters listed in L{_warnChars} 
    is also in the uri, an extra warning is also generated.
    @param uri: URI
    @param options: 
    @type options: L{Options<pyRdfa.Options>}
    """
    from . import err_unusual_char_in_URI
    suri = uri.strip()
    for c in _warnChars:
        if suri.find(c) != -1:
            if options != None:
                options.add_warning(err_unusual_char_in_URI % suri)
            break
    return quote(suri, _unquotedChars)
    
#########################################################################################################
    
def create_file_name(uri):
    """
    Create a suitable file name from an (absolute) URI. Used, eg, for the generation of a file name for a cached vocabulary file.
    """
    suri = uri.strip()
    final_uri = quote(suri,_unquotedChars)
    # Remove some potentially dangereous characters
    return final_uri.replace(' ','_').replace('%','_').replace('-','_').replace('+','_').replace('/','_').replace('?','_').replace(':','_').replace('=','_').replace('#','_')

#########################################################################################################
def has_one_of_attributes(node,*args):
    """
    Check whether one of the listed attributes is present on a (DOM) node.
    @param node: DOM element node
    @param args: possible attribute names
    @return: True or False
    @rtype: Boolean
    """
    if len(args) == 0:
        return None
    if isinstance(args[0], tuple) or isinstance(args[0], list):
        rargs = args[0]
    else:
        rargs = args
    
    return True in [ node.hasAttribute(attr) for attr in rargs ]

#########################################################################################################
def traverse_tree(node, func):
    """Traverse the whole element tree, and perform the function C{func} on all the elements.
    @param node: DOM element node
    @param func: function to be called on the node. Input parameter is a DOM Element Node. If the function returns a boolean True, the recursion is stopped.
    """
    if func(node):
        return

    for n in node.childNodes:
        if n.nodeType == node.ELEMENT_NODE:
            traverse_tree(n, func)

#########################################################################################################
def return_XML(state, inode, base=True, xmlns=True):
    """
    Get (recursively) the XML Literal content of a DOM Element Node. (Most of the processing is done
    via a C{node.toxml} call of the xml minidom implementation.)

    @param inode: DOM Node
    @param state: L{pyRdfa.state.ExecutionContext}
    @param base: whether the base element should be added to the output
    @type base: Boolean
    @param xmlns: whether the namespace declarations should be repeated in the generated node
    @type xmlns: Boolean
    @return: string
    """
    node = inode.cloneNode(True)
    # Decorate the element with namespaces value and, optionally, base
    if base:
        node.setAttribute("xml:base",state.base)
    if xmlns:
        for prefix in state.term_or_curie.xmlns:
            if not node.hasAttribute("xmlns:%s" % prefix):
                node.setAttribute("xmlns:%s" % prefix,"%s" % state.term_or_curie.xmlns[prefix])
        # Set the default namespace, if not done (and is available)
        if not node.getAttribute("xmlns") and state.defaultNS != None:
            node.setAttribute("xmlns", state.defaultNS)
    return node.toxml()

#########################################################################################################

def dump(node):
    """
    This is just for debug purposes: it prints the essential content of the node in the tree starting at node.

    @param node: DOM node
    """
    print(node.toprettyxml(indent="", newl=""))
    




# -*- coding: utf-8 -*-
"""
Implementation of the C{@property} value handling.

RDFa 1.0 and RDFa 1.1 are fairly different. RDFa 1.0 generates only literals, see
U{RDFa Task Force's wiki page<http://www.w3.org/2006/07/SWD/wiki/RDFa/LiteralObject>} for the details.
On the other hand, RDFa 1.1, beyond literals, can also generate URI references. Hence the duplicate method in the L{ProcessProperty} class, one for RDFa 1.0 and the other for RDFa 1.1.

@summary: RDFa Literal generation
@requires: U{RDFLib package<http://rdflib.net>}
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
"""

"""
$Id: property.py,v 1.14 2013-07-26 16:10:16 ivan Exp $
$Date: 2013-07-26 16:10:16 $
"""

from rdflib import BNode
from rdflib import Literal, URIRef
from rdflib import RDF as ns_rdf
from rdflib.term import XSDToPython

from . import IncorrectBlankNodeUsage, IncorrectLiteral, err_no_blank_node
from .utils import has_one_of_attributes, return_XML

import re

XMLLiteral  = ns_rdf["XMLLiteral"]
HTMLLiteral = URIRef("http://www.w3.org/1999/02/22-rdf-syntax-ns#HTML")

class ProcessProperty:
    """Generate the value for C{@property} taking into account datatype, etc.
    Note: this class is created only if the C{@property} is indeed present, no need to check.
    
    @ivar node: DOM element node
    @ivar graph: the (RDF) graph to add the properies to
    @ivar subject: the RDFLib URIRef serving as a subject for the generated triples
    @ivar state: the current state to be used for the CURIE-s
    @type state: L{state.ExecutionContext}
    @ivar typed_resource: Typically the bnode generated by a @typeof
    """
    def __init__(self, node, graph, subject, state, typed_resource=None):
        """
        @param node: DOM element node
        @param graph: the (RDF) graph to add the properies to
        @param subject: the RDFLib URIRef serving as a subject for the generated triples
        @param state: the current state to be used for the CURIE-s
        @param state: L{state.ExecutionContext}
        @param typed_resource: Typically the bnode generated by a @typeof; in RDFa 1.1, that becomes the object for C{@property}
        """
        self.node =           node
        self.graph =          graph
        self.subject =        subject
        self.state =          state
        self.typed_resource = typed_resource
        
    def generate(self):
        """
        Common entry point for the RDFa 1.0 and RDFa 1.1 versions; bifurcates based on the RDFa version, as retrieved from the state object.
        """
        if self.state.rdfa_version >= "1.1":
            self.generate_1_1()
        else:
            self.generate_1_0()

    def generate_1_1(self):
        """Generate the property object, 1.1 version"""

        #########################################################################        
        # See if the target is _not_ a literal
        irirefs =    ("resource", "href", "src")
        noiri =      ("content", "datatype", "rel", "rev")
        notypediri = ("content", "datatype", "rel", "rev", "about", "about_pruned")
        if has_one_of_attributes(self.node, irirefs) and not has_one_of_attributes(self.node, noiri):
            # @href/@resource/@src takes the lead here...
            obj = self.state.getResource(irirefs)
        elif self.node.hasAttribute("typeof") and not has_one_of_attributes(self.node, notypediri) and self.typed_resource != None:
                # a @typeof creates a special branch in case the typed resource was set during parsing
                obj = self.typed_resource
        else:
            # We have to generate a literal
            
            # Get, if exists, the value of @datatype
            datatype = ''
            dtset = False
            if self.node.hasAttribute("datatype"):
                dtset = True
                dt = self.node.getAttribute("datatype")
                if dt != "":
                    datatype = self.state.getURI("datatype")

            # Supress lange is set in case some elements explicitly want to supress the effect of language
            # There were discussions, for example, that the <time> element should do so. Although,
            # after all, this was reversed, the functionality is kept in the code in case another
            # element might need it...
            if self.state.lang != None and self.state.supress_lang == False:
                lang = self.state.lang
            else:
                lang = ''

            # The simple case: separate @content attribute
            if self.node.hasAttribute("content"):
                val = self.node.getAttribute("content")
                # Handling the automatic uri conversion case
                if dtset == False:
                    obj = Literal(val, lang=lang)
                else:
                    obj = self._create_Literal(val, datatype=datatype, lang=lang)
                # The value of datatype has been set, and the keyword paramaters take care of the rest
            else:
                # see if there *is* a datatype (even if it is empty!)
                if dtset:
                    if datatype == XMLLiteral:
                        litval = self._get_XML_literal(self.node)
                        obj = Literal(litval,datatype=XMLLiteral)
                    elif datatype == HTMLLiteral:
                        # I am not sure why this hack is necessary, but otherwise an encoding error occurs
                        # In Python3 all this should become moot, due to the unicode everywhere approach...
                        obj = Literal(self._get_HTML_literal(self.node), datatype=HTMLLiteral)
                    else:
                        obj = self._create_Literal(self._get_literal(self.node), datatype=datatype, lang=lang)
                else:
                    obj = self._create_Literal(self._get_literal(self.node), lang=lang)

        if obj != None:
            for prop in self.state.getURI("property"):
                if not isinstance(prop, BNode):
                    if self.node.hasAttribute("inlist"):
                        self.state.add_to_list_mapping(prop, obj)
                    else :            
                        self.graph.add( (self.subject, prop, obj) )
                else:
                    self.state.options.add_warning(err_no_blank_node % "property", warning_type=IncorrectBlankNodeUsage, node=self.node.nodeName)
    
        # return

    def generate_1_0(self):
        """Generate the property object, 1.0 version"""
                
        #########################################################################        
        # We have to generate a literal indeed.
        # Get, if exists, the value of @datatype
        datatype = ''
        dtset = False
        if self.node.hasAttribute("datatype"):
            dtset = True
            dt = self.node.getAttribute("datatype")
            if dt != "":
                datatype = self.state.getURI("datatype")
    
        if self.state.lang != None:
            lang = self.state.lang
        else:
            lang = ''

        # The simple case: separate @content attribute
        if self.node.hasAttribute("content"):
            val = self.node.getAttribute("content")
            # Handling the automatic uri conversion case
            if dtset == False:
                obj = Literal(val, lang=lang)
            else:
                obj = self._create_Literal(val, datatype=datatype, lang=lang)
            # The value of datatype has been set, and the keyword paramaters take care of the rest
        else:
            # see if there *is* a datatype (even if it is empty!)
            if dtset:
                # yep. The Literal content is the pure text part of the current element:
                # We have to check whether the specified datatype is, in fact, an
                # explicit XML Literal
                if datatype == XMLLiteral:
                    litval = self._get_XML_literal(self.node)
                    obj = Literal(litval,datatype=XMLLiteral)
                elif datatype == HTMLLiteral:
                    # I am not sure why this hack is necessary, but otherwise an encoding error occurs
                    # In Python3 all this should become moot, due to the unicode everywhere approach...
                    obj = Literal(self._get_HTML_literal(self.node), datatype=HTMLLiteral)
                else:
                    obj = self._create_Literal(self._get_literal(self.node), datatype=datatype, lang=lang)
            else:
                # no controlling @datatype. We have to see if there is markup in the contained
                # element
                if True in [ n.nodeType == self.node.ELEMENT_NODE for n in self.node.childNodes ]:
                    # yep, and XML Literal should be generated
                    obj = self._create_Literal(self._get_XML_literal(self.node), datatype=XMLLiteral)
                else:
                    # At this point, there might be entities in the string that are returned as real characters by the dom
                    # implementation. That should be turned back
                    obj = self._create_Literal(self._get_literal(self.node), lang=lang)
    
        for prop in self.state.getURI("property"):
            if not isinstance(prop,BNode):
                self.graph.add( (self.subject,prop,obj) )
            else:
                self.state.options.add_warning(err_no_blank_node % "property", warning_type=IncorrectBlankNodeUsage, node=self.node.nodeName)
    
        # return
    
    ######################################################################################################################################
    
    
    def _putBackEntities(self, data):
        """Put 'back' entities for the '&','<', and '>' characters, to produce a proper XML string.
        Used by the XML Literal extraction.
        @param data: string to be converted
        @return: string with entities
        @rtype: string
        """
        return data.replace('&','&amp;').replace('<','&lt;').replace('>','&gt;')
        
    def _get_literal(self, Pnode):
        """
        Get (recursively) the full text from a DOM Node.
    
        @param Pnode: DOM Node
        @return: string
        """
        rc = ""
        for node in Pnode.childNodes:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + node.data
            elif node.nodeType == node.ELEMENT_NODE:
                rc = rc + self._get_literal(node)
    
        # The decision of the group in February 2008 is not to normalize the result by default.
        # This is reflected in the default value of the option        
        
        if self.state.options.space_preserve:
            return rc
        else:
            return re.sub(r'(\r| |\n|\t)+'," ",rc).strip()
    # end getLiteral
    
    def _get_XML_literal(self, Pnode):
        """
        Get (recursively) the XML Literal content of a DOM Node. 
    
        @param Pnode: DOM Node
        @return: string
        """    
        rc = ""        
        for node in Pnode.childNodes:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + self._putBackEntities(node.data)
            elif node.nodeType == node.ELEMENT_NODE:
                rc = rc + return_XML(self.state, node, base = False)
        return rc
    # end getXMLLiteral

    def _get_HTML_literal(self, Pnode):
        """
        Get (recursively) the XML Literal content of a DOM Node. 
    
        @param Pnode: DOM Node
        @return: string
        """    
        rc = ""        
        for node in Pnode.childNodes:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + self._putBackEntities(node.data)
            elif node.nodeType == node.ELEMENT_NODE:
                rc = rc + return_XML(self.state, node, base = False, xmlns = False )
        return rc
    # end getHTMLLLiteral
    
    def _create_Literal(self, val, datatype = '', lang = ''):
        """
        Create a literal, taking into account the datatype and language.
        @return: Literal
        """
        if datatype == None or datatype == '':
            return Literal(val, lang=lang)
        #elif datatype == ns_xsd["string"]:
        #    return Literal(val)
        else:
            # This is a bit convoluted... the default setup of rdflib does not gracefully react if the
            # datatype cannot properly be converted to Python. I have to copy and reuse some of the
            # rdflib code to get this working...
            # To make things worse: rdlib 3.1.0 does not handle the various xsd date types properly, ie,
            # the conversion function below will generate errors. Ie, the check should be skipped for those
            convFunc = XSDToPython.get(datatype, None)
            if convFunc:
                try:
                    _pv = convFunc(val)
                    # If we got there the literal value and its datatype match
                except:
                    self.state.options.add_warning("Incompatible value (%s) and datatype (%s) in Literal definition." % (val, datatype), warning_type=IncorrectLiteral, node=self.node.nodeName)
            return Literal(val, datatype=datatype)



# -*- coding: utf-8 -*-
"""
The core parsing function of RDFa. Some details are
put into other modules to make it clearer to update/modify (e.g., generation of C{@property} values, or managing the current state).

Note that the entry point (L{parse_one_node}) bifurcates into an RDFa 1.0 and RDFa 1.1 version, ie,
to L{_parse_1_0} and L{_parse_1_1}. Some of the parsing details (management of C{@property}, list facilities, changed behavior on C{@typeof})) have changed
between versions and forcing the two into one function would be counter productive.

@summary: RDFa core parser processing step
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
"""

"""
$Id: parse.py,v 1.19 2013-01-07 12:46:43 ivan Exp $
$Date: 2013-01-07 12:46:43 $
"""

from .state import ExecutionContext
from .property import ProcessProperty
from .embeddedRDF import handle_embeddedRDF
from .host import HostLanguage, host_dom_transforms

from rdflib import URIRef
from rdflib import BNode
from rdflib import RDF as ns_rdf

from . import IncorrectBlankNodeUsage, err_no_blank_node
from .utils import has_one_of_attributes

#######################################################################
def parse_one_node(node, graph, parent_object, incoming_state, parent_incomplete_triples):
    """The (recursive) step of handling a single node. 
    
    This entry just switches between the RDFa 1.0 and RDFa 1.1 versions for parsing. This method is only invoked once,
    actually, from the top level; the recursion then happens in the L{_parse_1_0} and L{_parse_1_1} methods for
    RDFa 1.0 and RDFa 1.1, respectively.

    @param node: the DOM node to handle
    @param graph: the RDF graph
    @type graph: RDFLib's Graph object instance
    @param parent_object: the parent's object, as an RDFLib URIRef
    @param incoming_state: the inherited state (namespaces, lang, etc.)
    @type incoming_state: L{state.ExecutionContext}
    @param parent_incomplete_triples: list of hanging triples (the missing resource set to None) to be handled (or not)
    by the current node.
    @return: whether the caller has to complete it's parent's incomplete triples
    @rtype: Boolean
    """
    # Branch according to versions.
    if incoming_state.rdfa_version >= "1.1":
        _parse_1_1(node, graph, parent_object, incoming_state, parent_incomplete_triples)
    else:
        _parse_1_0(node, graph, parent_object, incoming_state, parent_incomplete_triples)

#######################################################################
def _parse_1_1(node, graph, parent_object, incoming_state, parent_incomplete_triples):
    """The (recursive) step of handling a single node. See the
    U{RDFa 1.1 Core document<http://www.w3.org/TR/rdfa-core/>} for further details.
    
    This is the RDFa 1.1 version.

    @param node: the DOM node to handle
    @param graph: the RDF graph
    @type graph: RDFLib's Graph object instance
    @param parent_object: the parent's object, as an RDFLib URIRef
    @param incoming_state: the inherited state (namespaces, lang, etc.)
    @type incoming_state: L{state.ExecutionContext}
    @param parent_incomplete_triples: list of hanging triples (the missing resource set to None) to be handled (or not)
    by the current node.
    @return: whether the caller has to complete it's parent's incomplete triples
    @rtype: Boolean
    """
    def header_check(p_obj):
        """Special disposition for the HTML <head> and <body> elements..."""
        if state.options.host_language in [ HostLanguage.xhtml, HostLanguage.html5, HostLanguage.xhtml5 ]:
            if node.nodeName == "head" or node.nodeName == "body":
                if not has_one_of_attributes(node, "about", "resource", "src", "href"):
                    return p_obj
        else:
            return None

    def lite_check():
        if state.options.check_lite and state.options.host_language in [ HostLanguage.html5, HostLanguage.xhtml5, HostLanguage.xhtml ]:
            if node.tagName == "link" and node.hasAttribute("rel") and state.term_or_curie.CURIE_to_URI(node.getAttribute("rel")) != None:
                state.options.add_warning("In RDFa Lite, attribute @rel in <link> is only used in non-RDFa way (consider using @property)", node=node)

    # Update the state. This means, for example, the possible local settings of
    # namespaces and lang
    state = ExecutionContext(node, graph, inherited_state=incoming_state)

    #---------------------------------------------------------------------------------
    # Extra warning check on RDFa Lite
    lite_check()
    
    #---------------------------------------------------------------------------------
    # Handling the role attribute is pretty much orthogonal to everything else...
    handle_role_attribute(node, graph, state)

    #---------------------------------------------------------------------------------
    # Handle the special case for embedded RDF, eg, in SVG1.2. 
    # This may add some triples to the target graph that does not originate from RDFa parsing
    # If the function return TRUE, that means that an rdf:RDF has been found. No
    # RDFa parsing should be done on that subtree, so we simply return...
    if state.options.embedded_rdf and node.nodeType == node.ELEMENT_NODE and handle_embeddedRDF(node, graph, state) : 
        return    

    #---------------------------------------------------------------------------------
    # calling the host language specific massaging of the DOM
    if state.options.host_language in host_dom_transforms and node.nodeType == node.ELEMENT_NODE:
        for func in host_dom_transforms[state.options.host_language] : func(node, state)

    #---------------------------------------------------------------------------------
    # First, let us check whether there is anything to do at all. Ie,
    # whether there is any relevant RDFa specific attribute on the element
    #
    if not has_one_of_attributes(node, "href", "resource", "about", "property", "rel", "rev", "typeof", "src", "vocab", "prefix"):
        # nop, there is nothing to do here, just go down the tree and return...
        for n in node.childNodes:
            if n.nodeType == node.ELEMENT_NODE : parse_one_node(n, graph, parent_object, state, parent_incomplete_triples)
        return

    #-----------------------------------------------------------------
    # The goal is to establish the subject and object for local processing
    # The behaviour is slightly different depending on the presense or not
    # of the @rel/@rev attributes
    current_subject = None
    current_object  = None
    typed_resource    = None
    
    if has_one_of_attributes(node, "rel", "rev") :
        # in this case there is the notion of 'left' and 'right' of @rel/@rev
        # in establishing the new Subject and the objectResource
        current_subject = header_check(parent_object)

        # set first the subject
        if node.hasAttribute("about"):
            current_subject = state.getURI("about")
            if node.hasAttribute("typeof") : typed_resource = current_subject
            
        # get_URI may return None in case of an illegal CURIE, so
        # we have to be careful here, not use only an 'else'
        if current_subject == None:
            current_subject = parent_object
        else:
            state.reset_list_mapping(origin = current_subject)
        
        # set the object resource
        current_object = state.getResource("resource", "href", "src")
            
        if node.hasAttribute("typeof") and not node.hasAttribute("about"):
            if current_object == None:
                current_object = BNode()
            typed_resource = current_object
        
        if not node.hasAttribute("inlist") and current_object != None:
            # In this case the newly defined object is, in fact, the head of the list
            # just reset the whole thing.
            state.reset_list_mapping(origin = current_object)

    elif  node.hasAttribute("property") and not has_one_of_attributes(node, "content", "datatype"):
        current_subject = header_check(parent_object)

        # this is the case when the property may take hold of @src and friends...
        if node.hasAttribute("about"):
            current_subject = state.getURI("about")
            if node.hasAttribute("typeof") : typed_resource = current_subject

        # getURI may return None in case of an illegal CURIE, so
        # we have to be careful here, not use only an 'else'
        if current_subject == None:
            current_subject = parent_object
        else:
            state.reset_list_mapping(origin = current_subject)

        if typed_resource == None and node.hasAttribute("typeof"):
            typed_resource = state.getResource("resource", "href", "src")
            if typed_resource == None:
                typed_resource = BNode()
            current_object = typed_resource
        else:
            current_object = current_subject
            
    else:
        current_subject = header_check(parent_object)

        # in this case all the various 'resource' setting attributes
        # behave identically, though they also have their own priority
        if current_subject == None:
            current_subject = state.getResource("about", "resource", "href", "src")
            
        # get_URI_ref may return None in case of an illegal CURIE, so
        # we have to be careful here, not use only an 'else'
        if current_subject == None:
            if node.hasAttribute("typeof"):
                current_subject = BNode()
                state.reset_list_mapping(origin = current_subject)
            else:
                current_subject = parent_object
        else:
            state.reset_list_mapping(origin = current_subject)

        # in this case no non-literal triples will be generated, so the
        # only role of the current_object Resource is to be transferred to
        # the children node
        current_object = current_subject
        if node.hasAttribute("typeof") : typed_resource = current_subject
        
    # ---------------------------------------------------------------------
    ## The possible typeof indicates a number of type statements on the typed resource
    for defined_type in state.getURI("typeof"):
        if typed_resource:
            graph.add((typed_resource, ns_rdf["type"], defined_type))

    # ---------------------------------------------------------------------
    # In case of @rel/@rev, either triples or incomplete triples are generated
    # the (possible) incomplete triples are collected, to be forwarded to the children
    incomplete_triples  = []
    for prop in state.getURI("rel"):
        if not isinstance(prop,BNode):
            if node.hasAttribute("inlist"):
                if current_object != None:
                    # Add the content to the list. Note that if the same list
                    # was initialized, at some point, by a None, it will be
                    # overwritten by this real content
                    state.add_to_list_mapping(prop, current_object)
                else:
                    # Add a dummy entry to the list... Note that
                    # if that list was initialized already with a real content
                    # this call will have no effect
                    state.add_to_list_mapping(prop, None)
                    
                    # Add a placeholder into the hanging rels
                    incomplete_triples.append( (None, prop, None) )
            else:
                theTriple = (current_subject, prop, current_object)
                if current_object != None:
                    graph.add(theTriple)
                else:
                    incomplete_triples.append(theTriple)
        else:
            state.options.add_warning(err_no_blank_node % "rel", warning_type=IncorrectBlankNodeUsage, node=node.nodeName)

    for prop in state.getURI("rev"):
        if not isinstance(prop,BNode):
            theTriple = (current_object,prop,current_subject)
            if current_object != None:
                graph.add(theTriple)
            else:
                incomplete_triples.append(theTriple)
        else:
            state.options.add_warning(err_no_blank_node % "rev", warning_type=IncorrectBlankNodeUsage, node=node.nodeName)

    # ----------------------------------------------------------------------
    # Generation of the @property values, including literals. The newSubject is the subject
    # A particularity of property is that it stops the parsing down the DOM tree if an XML Literal is generated,
    # because everything down there is part of the generated literal. 
    if node.hasAttribute("property"):
        ProcessProperty(node, graph, current_subject, state, typed_resource).generate_1_1()

    # ----------------------------------------------------------------------
    # Setting the current object to a bnode is setting up a possible resource
    # for the incomplete triples downwards
    if current_object == None:
        object_to_children = BNode()
    else:
        object_to_children = current_object

    #-----------------------------------------------------------------------
    # Here is the recursion step for all the children
    for n in node.childNodes:
        if n.nodeType == node.ELEMENT_NODE : 
            _parse_1_1(n, graph, object_to_children, state, incomplete_triples)

    # ---------------------------------------------------------------------
    # At this point, the parent's incomplete triples may be completed
    for (s,p,o) in parent_incomplete_triples:
        if s == None and o == None:
            # This is an encoded version of a hanging rel for a collection:
            incoming_state.add_to_list_mapping( p, current_subject )
        else:
            if s == None : s = current_subject
            if o == None : o = current_subject
            graph.add((s,p,o))

    # Generate the lists, if any and if this is the level where a new list was originally created    
    if state.new_list and not state.list_empty():
        for prop in state.get_list_props():
            vals  = state.get_list_value(prop)
            if vals == None:
                # This was an empty list, in fact, ie, the list has been initiated by a <xxx rel="prop" inlist>
                # but no list content has ever been added
                graph.add( (state.get_list_origin(), prop, ns_rdf["nil"]) )
            else:
                heads = [ BNode() for _r in vals ] + [ ns_rdf["nil"] ]
                for i in range(0, len(vals)):
                    graph.add( (heads[i], ns_rdf["first"], vals[i]) )
                    graph.add( (heads[i], ns_rdf["rest"],  heads[i+1]) )
                # Anchor the list
                graph.add( (state.get_list_origin(), prop, heads[0]) )

    # -------------------------------------------------------------------
    # This should be it...
    # -------------------------------------------------------------------
    return


##################################################################################################################
def _parse_1_0(node, graph, parent_object, incoming_state, parent_incomplete_triples):
    """The (recursive) step of handling a single node. See the
    U{RDFa 1.0 syntax document<http://www.w3.org/TR/rdfa-syntax>} for further details.
    
    This is the RDFa 1.0 version.

    @param node: the DOM node to handle
    @param graph: the RDF graph
    @type graph: RDFLib's Graph object instance
    @param parent_object: the parent's object, as an RDFLib URIRef
    @param incoming_state: the inherited state (namespaces, lang, etc.)
    @type incoming_state: L{state.ExecutionContext}
    @param parent_incomplete_triples: list of hanging triples (the missing resource set to None) to be handled (or not)
    by the current node.
    @return: whether the caller has to complete it's parent's incomplete triples
    @rtype: Boolean
    """

    # Update the state. This means, for example, the possible local settings of
    # namespaces and lang
    state = ExecutionContext(node, graph, inherited_state=incoming_state)

    #---------------------------------------------------------------------------------
    # Handling the role attribute is pretty much orthogonal to everything else...
    handle_role_attribute(node, graph, state)

    #---------------------------------------------------------------------------------
    # Handle the special case for embedded RDF, eg, in SVG1.2. 
    # This may add some triples to the target graph that does not originate from RDFa parsing
    # If the function return TRUE, that means that an rdf:RDF has been found. No
    # RDFa parsing should be done on that subtree, so we simply return...
    if state.options.embedded_rdf and node.nodeType == node.ELEMENT_NODE and handle_embeddedRDF(node, graph, state) : 
        return    

    #---------------------------------------------------------------------------------
    # calling the host language specific massaging of the DOM
    if state.options.host_language in host_dom_transforms and node.nodeType == node.ELEMENT_NODE:
        for func in host_dom_transforms[state.options.host_language] : func(node, state)

    #---------------------------------------------------------------------------------
    # First, let us check whether there is anything to do at all. Ie,
    # whether there is any relevant RDFa specific attribute on the element
    #
    if not has_one_of_attributes(node, "href", "resource", "about", "property", "rel", "rev", "typeof", "src"):
        # nop, there is nothing to do here, just go down the tree and return...
        for n in node.childNodes:
            if n.nodeType == node.ELEMENT_NODE : parse_one_node(n, graph, parent_object, state, parent_incomplete_triples)
        return

    #-----------------------------------------------------------------
    # The goal is to establish the subject and object for local processing
    # The behaviour is slightly different depending on the presense or not
    # of the @rel/@rev attributes
    current_subject = None
    current_object = None

    if has_one_of_attributes(node, "rel", "rev") :
        # in this case there is the notion of 'left' and 'right' of @rel/@rev
        # in establishing the new Subject and the objectResource
        current_subject = state.getResource("about","src")

        # get_URI may return None in case of an illegal CURIE, so
        # we have to be careful here, not use only an 'else'
        if current_subject == None:
            if node.hasAttribute("typeof"):
                current_subject = BNode()
            else:
                current_subject = parent_object
        else:
            state.reset_list_mapping(origin = current_subject)
        
        # set the object resource
        current_object = state.getResource("resource", "href")
        
    else:
        # in this case all the various 'resource' setting attributes
        # behave identically, though they also have their own priority
        current_subject = state.getResource("about", "src", "resource", "href")
        
        # get_URI_ref may return None in case of an illegal CURIE, so
        # we have to be careful here, not use only an 'else'
        if current_subject == None:
            if node.hasAttribute("typeof"):
                current_subject = BNode()
            else:
                current_subject = parent_object
            current_subject = parent_object
        else:
            state.reset_list_mapping(origin = current_subject)

        # in this case no non-literal triples will be generated, so the
        # only role of the current_object Resource is to be transferred to
        # the children node
        current_object = current_subject

    # ---------------------------------------------------------------------
    ## The possible typeof indicates a number of type statements on the new Subject
    for defined_type in state.getURI("typeof"):
        graph.add((current_subject, ns_rdf["type"], defined_type))

    # ---------------------------------------------------------------------
    # In case of @rel/@rev, either triples or incomplete triples are generated
    # the (possible) incomplete triples are collected, to be forwarded to the children
    incomplete_triples  = []
    for prop in state.getURI("rel"):
        if not isinstance(prop,BNode):
            theTriple = (current_subject, prop, current_object)
            if current_object != None:
                graph.add(theTriple)
            else:
                incomplete_triples.append(theTriple)
        else:
            state.options.add_warning(err_no_blank_node % "rel", warning_type=IncorrectBlankNodeUsage, node=node.nodeName)

    for prop in state.getURI("rev"):
        if not isinstance(prop,BNode):
            theTriple = (current_object,prop,current_subject)
            if current_object != None:
                graph.add(theTriple)
            else:
                incomplete_triples.append(theTriple)
        else:
            state.options.add_warning(err_no_blank_node % "rev", warning_type=IncorrectBlankNodeUsage, node=node.nodeName)

    # ----------------------------------------------------------------------
    # Generation of the literal values. The newSubject is the subject
    # A particularity of property is that it stops the parsing down the DOM tree if an XML Literal is generated,
    # because everything down there is part of the generated literal. 
    if node.hasAttribute("property"):
        ProcessProperty(node, graph, current_subject, state).generate_1_0()

    # ----------------------------------------------------------------------
    # Setting the current object to a bnode is setting up a possible resource
    # for the incomplete triples downwards
    if current_object == None:
        object_to_children = BNode()
    else:
        object_to_children = current_object

    #-----------------------------------------------------------------------
    # Here is the recursion step for all the children
    for n in node.childNodes:
        if n.nodeType == node.ELEMENT_NODE : 
            _parse_1_0(n, graph, object_to_children, state, incomplete_triples)

    # ---------------------------------------------------------------------
    # At this point, the parent's incomplete triples may be completed
    for (s,p,o) in parent_incomplete_triples:
        if s == None and o == None:
            # This is an encoded version of a hanging rel for a collection:
            incoming_state.add_to_list_mapping( p, current_subject )
        else:
            if s == None : s = current_subject
            if o == None : o = current_subject
            graph.add((s,p,o))

    # -------------------------------------------------------------------
    # This should be it...
    # -------------------------------------------------------------------
    return


#######################################################################
# Handle the role attribute
def handle_role_attribute(node, graph, state):
    """
    Handling the role attribute, according to http://www.w3.org/TR/role-attribute/#using-role-in-conjunction-with-rdfa
    @param node: the DOM node to handle
    @param graph: the RDF graph
    @type graph: RDFLib's Graph object instance
    @param state: the inherited state (namespaces, lang, etc.)
    @type state: L{state.ExecutionContext}
    """
    if node.hasAttribute("role"):
        if node.hasAttribute("id"):
            i = node.getAttribute("id").strip()
            subject = URIRef(state.base + '#' + i)
        else:
            subject = BNode()
        predicate = URIRef('http://www.w3.org/1999/xhtml/vocab#role')
        for obj in state.getURI("role"):
            graph.add((subject, predicate, obj))




# -*- coding: utf-8 -*-
"""
Transformer sub-package for the pyRdfa package. It contains modules with transformer functions; each may be
invoked by pyRdfa to transform the dom tree before the "real" RDfa processing.

@summary: RDFa Transformer package
@requires: U{RDFLib package<http://rdflib.net>}
@organization: U{World Wide Web Consortium<http://www.w3.org>}
@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
"""

"""
$Id: __init__.py,v 1.8 2012/06/12 11:47:19 ivan Exp $
$Date: 2012/06/12 11:47:19 $
"""
__version__ = "3.0"

# Here are the transfomer functions that are to be performed for all RDFa files, no matter what

def top_about(root, options, state):
    """
    @param root: a DOM node for the top level element
    @param options: invocation options
    @type options: L{Options<pyRdfa.options>}
    @param state: top level execution state
    @type state: L{State<pyRdfa.state>}
    """
    def set_about(node):
        if has_one_of_attributes(node, "rel", "rev"):
            if not has_one_of_attributes(top, "about", "src"):
                node.setAttribute("about","")
        else:
            if not has_one_of_attributes(node, "href", "resource", "about", "src"):
                node.setAttribute("about","")
    
    from ..host import HostLanguage
    from ..utils import has_one_of_attributes
        
    if not has_one_of_attributes(root, "about"):
        # The situation is a bit complicated: if a @resource is present without anything else, then it sets
        # the subject, ie, should be accepted...
        if has_one_of_attributes(root, "resource", "href", "src"):
            if has_one_of_attributes(root, "rel", "rev","property"):
                root.setAttribute("about","")
        else:
            root.setAttribute("about","")
        
    if options.host_language in [ HostLanguage.xhtml, HostLanguage.html5, HostLanguage.xhtml5 ]:
        if state.rdfa_version >= "1.1":
            pass
        else:
            for top in root.getElementsByTagName("head"):
                if not has_one_of_attributes(top, "href", "resource", "about", "src"):
                    set_about(top)
            for top in root.getElementsByTagName("body"):
                if not has_one_of_attributes(top, "href", "resource", "about", "src"):
                    set_about(top)


def empty_safe_curie(node, options, state):
    """
    Remove the attributes whose value is an empty safe curie. It also adds an 'artificial' flag, ie, an
    attribute (called 'emptysc') into the node to signal that there _is_ an attribute with an ignored
    safe curie value. The name of the attribute is 'about_pruned' or 'resource_pruned'.
    
    @param node: a DOM node for the top level element
    @param options: invocation options
    @type options: L{Options<pyRdfa.options>}
    @param state: top level execution state
    @type state: L{State<pyRdfa.state>}
    """
    def prune_safe_curie(node,name):
        if node.hasAttribute(name):
            av = node.getAttribute(name)
            if av == '[]':
                node.removeAttribute(name)
                node.setAttribute(name+'_pruned','')
                msg = "Attribute @%s uses an empty safe CURIE; the attribute is ignored" % name
                options.add_warning(msg, node=node)
                
    prune_safe_curie(node, "about")
    prune_safe_curie(node, "resource")
    for n in node.childNodes:
        if n.nodeType == node.ELEMENT_NODE:
            empty_safe_curie(n, options, state)

def vocab_for_role(node, options, state):
    """
    The value of the @role attribute (defined separately in the U{Role Attribute Specification Lite<http://www.w3.org/TR/role-attribute/#using-role-in-conjunction-with-rdfa>}) should be as if a @vocab value to the
    XHTML vocabulary was defined for it. This method turns all terms in role attributes into full URI-s, so that
    this would not be an issue for the run-time.
    
    @param node: a DOM node for the top level element
    @param options: invocation options
    @type options: L{Options<pyRdfa.options>}
    @param state: top level execution state
    @type state: L{State<pyRdfa.state>}
    """
    from ..termorcurie import termname, XHTML_URI

    def handle_role(node):
        if node.hasAttribute("role"):
            old_values = node.getAttribute("role").strip().split()
            new_values = ""
            for val in old_values:
                if termname.match(val):
                    new_values += XHTML_URI + val + ' '
                else:
                    new_values += val + ' '
            node.setAttribute("role", new_values.strip())

    handle_role(node)
    for n in node.childNodes:
        if n.nodeType == node.ELEMENT_NODE:
            vocab_for_role(n, options, state)




# -*- coding: utf-8 -*-
"""
Extracting possible embedded RDF/XML content from the file and parse it separately into the Graph. This is used, for example
by U{SVG 1.2 Tiny<http://www.w3.org/TR/SVGMobile12/>}.

@author: U{Ivan Herman<a href="http://www.w3.org/People/Ivan/">}
@license: This software is available for use under the
U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231">}
@contact: Ivan Herman, ivan@w3.org
@version: $Id: embeddedRDF.py,v 1.15 2012/11/16 17:51:53 ivan Exp $
"""

from io import StringIO

from .host import accept_embedded_rdf_xml, accept_embedded_turtle
from .utils import return_XML
import sys

def handle_embeddedRDF(node, graph, state):
    """
    Handles embedded RDF. There are two possibilities:
    
     - the file is one of the XML dialects that allows for an embedded RDF/XML portion. See the L{host.accept_embedded_rdf_xml} for those (a typical example is SVG). 
     - the file is HTML and there is a turtle portion in the C{<script>} element with type text/turtle. 
    
    @param node: a DOM node for the top level element
    @param graph: target rdf graph
    @type graph: RDFLib's Graph object instance
    @param state: the inherited state (namespaces, lang, etc)
    @type state: L{state.ExecutionContext}
    @return: whether an RDF/XML or turtle content has been detected or not. If TRUE, the RDFa processing should not occur on the node and its descendents. 
    @rtype: Boolean
    """
    #def _get_prefixes_in_turtle():
    #    retval = ""
    #    for key in state.term_or_curie.ns:
    #        retval += "@prefix %s: <%s> .\n" % (key, state.term_or_curie.ns[key])
    #    retval += '\n'
    #    return retval
    
    # This feature is optional!
    def _get_literal(Pnode):
        """
        Get the full text
        @param Pnode: DOM Node
        @return: string
        """
        rc = ""
        for node in Pnode.childNodes:
            if node.nodeType in [node.TEXT_NODE, node.CDATA_SECTION_NODE]:
                rc = rc + node.data
        # Sigh... the HTML5 parser does not recognize the CDATA escapes, ie, it just passes on the <![CDATA[ and ]]> strings:-(
        return rc.replace("<![CDATA[","").replace("]]>","")

    if state.options.embedded_rdf:
        # Embedded turtle, per the latest Turtle draft
        if state.options.host_language in accept_embedded_turtle and node.nodeName.lower() == "script":
            if node.hasAttribute("type") and node.getAttribute("type") == "text/turtle":
                #prefixes = _get_prefixes_in_turtle()
                #content  = _get_literal(node)
                #rdf = StringIO(prefixes + content)
                content  = _get_literal(node)
                rdf = StringIO(content)
                try:
                    graph.parse(rdf, format="n3", publicID = state.base)
                    state.options.add_info("The output graph includes triples coming from an embedded Turtle script")
                except:
                    _type, value, _traceback = sys.exc_info()
                    state.options.add_error("Embedded Turtle content could not be parsed (problems with %s?); ignored" % value)
            return True
        elif state.options.host_language in accept_embedded_rdf_xml and node.localName == "RDF" and node.namespaceURI == "http://www.w3.org/1999/02/22-rdf-syntax-ns#":
            rdf = StringIO(return_XML(state, node))
            try:
                graph.parse(rdf)
                state.options.add_info("The output graph includes triples coming from an embedded RDF/XML subtree")
            except:
                _type, value, _traceback = sys.exc_info()
                state.options.add_error("Embedded RDF/XML content could not parsed (problems with %s?); ignored" % value)
            return True
        else:
            return False
    else:
        return False




# -*- coding: utf-8 -*-
"""
`Options` class collect the possible options that govern the parsing possibilities. The module also includes the `ProcessorGraph` class that handles the processor graph, per RDFa 1.1 (i.e., the graph containing errors and warnings). 
"""

import datetime

from rdflib import URIRef
from rdflib import Literal
from rdflib import BNode
from rdflib import Namespace
from rdflib import Graph
from rdflib import RDF as ns_rdf

from .host import HostLanguage, content_to_host_language, predefined_1_0_rel, require_embedded_rdf
from . import ns_xsd, ns_distill, ns_rdfa
from . import RDFA_Error, RDFA_Warning, RDFA_Info
from .transform.lite import lite_prune

ns_dc = Namespace("http://purl.org/dc/terms/")
ns_ht = Namespace("http://www.w3.org/2006/http#")

class ProcessorGraph:
    """Wrapper around the 'processor graph', ie, the (RDF) Graph containing the warnings,
    error messages, and informational messages.
    """
    def __init__(self):
        self.graph = Graph()
        
    def add_triples(self, msg, top_class, info_class, context, node):
        """
        Add an error structure to the processor graph: a bnode with a number of predicates. The structure
        follows U{the processor graph vocabulary<http://www.w3.org/2010/02/rdfa/wiki/Processor_Graph_Vocabulary>} as described
        on the RDFa WG Wiki page.
        
        @param msg: the core error message, added as an object to a dc:description
        @param top_class: Error, Warning, or Info; an explicit rdf:type added to the bnode
        @type top_class: URIRef
        @param info_class: An additional error class, added as an rdf:type to the bnode in case it is not None
        @type info_class: URIRef
        @param context: An additional information added, if not None, as an object with rdfa:context as a predicate
        @type context: either an URIRef or a URI String (an URIRef will be created in the second case)
        @param node: The node's element name that contains the error
        @type node: string
        @return: the bnode that serves as a subject for the errors. The caller may add additional information
        @rtype: BNode
        """
        # Lazy binding of relevant prefixes
        self.graph.bind("dcterms", ns_dc)
        self.graph.bind("pyrdfa",  ns_distill)
        self.graph.bind("rdf",     ns_rdf)
        self.graph.bind("rdfa",    ns_rdfa)
        self.graph.bind("ht",      ns_ht)
        self.graph.bind("xsd",     ns_xsd)

        is_context_string = isinstance(context, str)

        bnode = BNode()

        if node != None:
            try:
                full_msg = "[In element '%s'] %s" % (node.nodeName, msg)
            except:
                full_msg = "[In element '%s'] %s" % (node, msg)
        else:
            full_msg = msg

        self.graph.add((bnode, ns_rdf["type"], top_class))
        if info_class:
            self.graph.add((bnode, ns_rdf["type"], info_class))
        self.graph.add((bnode, ns_dc["description"], Literal(full_msg)))
        self.graph.add((bnode, ns_dc["date"], Literal(datetime.datetime.utcnow().isoformat(),datatype=ns_xsd["dateTime"])))
        if context and (isinstance(context,URIRef) or is_context_string):
            htbnode = BNode()
            self.graph.add( (bnode,   ns_rdfa["context"],htbnode) )
            self.graph.add( (htbnode, ns_rdf["type"], ns_ht["Request"]) )
            self.graph.add( (htbnode, ns_ht["requestURI"], Literal("%s" % context)) )
        return bnode
    
    def add_http_context(self, subj, http_code):
        """
        Add an additional HTTP context to a message with subject in C{subj}, using the U{<http://www.w3.org/2006/http#>}
        vocabulary. Typically used to extend an error structure, as created by L{add_triples}.
        
        @param subj: an RDFLib resource, typically a blank node
        @param http_code: HTTP status code
        """
        bnode = BNode()
        self.graph.add((subj, ns_rdfa["context"], bnode))
        self.graph.add((bnode, ns_rdf["type"], ns_ht["Response"]))
        self.graph.add((bnode, ns_ht["responseCode"], URIRef("http://www.w3.org/2006/http#%s" % http_code)))

class Options:
    '''Settable options. An instance of this class is stored in the `ExecutionContext` of the parser.'''

    def __init__(self, output_default_graph =       True,
                       output_processor_graph =     False,
                       space_preserve =             True,
                       transformers =               [],
                       embedded_rdf =               True,
                       vocab_expansion =            False,
                       vocab_cache =                True,
                       vocab_cache_report =         False,
                       refresh_vocab_cache =        False,
                       add_informational_messages = False,
                       check_lite =                 False,
                       experimental_features =      False,
                       certifi_verify =             True):

        self.space_preserve =             space_preserve
        '''whether plain literals should preserve spaces at output or not'''
        self.transformers =               transformers
        '''extra transformers'''
        self.processor_graph =            ProcessorGraph()
        '''the `ProcessorGraph`'''
        self.output_default_graph =       output_default_graph
        '''whether the 'default' graph should be returned to the user'''
        self.output_processor_graph =     output_processor_graph
        '''whether the 'processor' graph should be returned to the user'''
        self.host_language =              HostLanguage.rdfa_core
        '''the host language for the RDFa attributes. Default is HostLanguage.xhtml, but it can be HostLanguage.rdfa_core and HostLanguage.html5, or others...'''
        self.vocab_cache_report =         vocab_cache_report
        '''whether the details of vocabulary file caching process should be reported as information (mainly for debug)'''
        self.refresh_vocab_cache =        refresh_vocab_cache
        '''whether the caching checks of vocabs should be by-passed, ie, if caches should be re-generated regardless of the stored date (important for vocab development)'''
        self.embedded_rdf =               embedded_rdf
        '''whether embedded RDF (ie, turtle in an HTML script element or an RDF/XML content in SVG) should be extracted and added to the final graph. This is a non-standard option...'''
        self.vocab_expansion =            vocab_expansion
        '''whether the @vocab elements should be expanded and a mini-RDFS processing should be done on the merged graph'''
        self.vocab_cache =                vocab_cache
        '''whether the system should use the vocabulary caching mechanism when expanding via the mini-RDFS, or should just fetch the graphs every time'''
        self.add_informational_messages = add_informational_messages
        '''whether informational messages should also be added to the processor graph, or only errors and warnings'''
        self.check_lite =                 check_lite
        '''whether RDFa Lite should be checked, to generate warnings'''
        self.experimental_features =      experimental_features
        '''whether experimental features should be activated; that is a developer's option...'''
        self.certifi_verify =             certifi_verify
        '''whether the SSL certificate should be verified'''
        if check_lite:
            self.transformers.append(lite_prune)

    def set_host_language(self, content_type):
        """
        Set the host language for processing, based on the recognized types. If this is not a recognized content type,
        it falls back to RDFa core (i.e., XML)
        @param content_type: content type
        @type content_type: string
        """
        if content_type in content_to_host_language:
            self.host_language = content_to_host_language[content_type]
        else:
            self.host_language = HostLanguage.rdfa_core
            
        if self.host_language in require_embedded_rdf:
            self.embedded_rdf = True

    def __str__(self):
        retval = """Current options:
        preserve space                         : %s
        output processor graph                 : %s
        output default graph                   : %s
        host language                          : %s
        accept embedded RDF                    : %s
        perfom semantic postprocessing         : %s
        cache vocabulary graphs                : %s
        """
        return retval % (self.space_preserve, self.output_processor_graph, self.output_default_graph, self.host_language, self.embedded_rdf, self.rdfa_sem, self.vocab_cache)

    def reset_processor_graph(self):
        """Empty the processor graph. This is necessary if the same options is reused
        for several RDFa sources, and new error messages should be generated.
        """
        self.processor_graph.graph.remove((None,None,None))

    def add_warning(self, txt, warning_type=None, context=None, node=None, buggy_value=None):
        """Add a warning to the processor graph.
        @param txt: the warning text. 
        @keyword warning_type: Warning Class
        @type warning_type: URIRef
        @keyword context: possible context to be added to the processor graph
        @type context: URIRef or String
        @keyword buggy_value: a special case when a 'term' is not recognized; no warning is generated for that case if the value is part of the 'usual' XHTML terms, because almost all RDFa file contains some of those and that would pollute the output
        @type buggy_value: String
        """
        if warning_type == ns_rdfa["UnresolvedTerm"] and buggy_value in predefined_1_0_rel:
            return
        return self.processor_graph.add_triples(txt, RDFA_Warning, warning_type, context, node)

    def add_info(self, txt, info_type=None, context=None, node=None, buggy_value=None):
        """Add an informational comment to the processor graph.
        @param txt: the information text. 
        @keyword info_type: Info Class
        @type info_type: URIRef
        @keyword context: possible context to be added to the processor graph
        @type context: URIRef or String
        @keyword buggy_value: a special case when a 'term' is not recognized; no information is generated for that case if the value is part of the 'usual' XHTML terms, because almost all RDFa file contains some of those and that would pollute the output
        @type buggy_value: String
        """
        if self.add_informational_messages:
            return self.processor_graph.add_triples(txt, RDFA_Info, info_type, context, node)
        else:
            return

    def add_error(self, txt, err_type=None, context=None, node=None, buggy_value=None):
        """Add an error  to the processor graph.
        @param txt: the information text. 
        @keyword err_type: Error Class
        @type err_type: URIRef
        @keyword context: possible context to be added to the processor graph
        @type context: URIRef or String
        @keyword buggy_value: a special case when a 'term' is not recognized; no error is generated for that case if the value is part of the 'usual' XHTML terms, because almost all RDFa file contains some of those and that would pollute the output
        @type buggy_value: String
        """
        return self.processor_graph.add_triples(txt, RDFA_Error, err_type, context, node)

